from typing import NamedTuple

import geopandas as gpd
import numba_celltree
import numpy as np
import pyproj
import pytest
import shapely
import xarray as xr
from matplotlib.collections import LineCollection
from scipy import sparse, spatial

import xugrid

try:
    import meshkernel as mk
except ImportError:
    pass

from . import requires_meshkernel

NAME = "mesh2d"
VERTICES = np.array(
    [
        [0.0, 0.0],  # 0
        [1.0, 0.0],  # 1
        [2.0, 0.0],  # 2
        [0.0, 1.0],  # 3
        [1.0, 1.0],  # 4
        [2.0, 1.0],  # 5
        [1.0, 2.0],  # 6
    ]
)
FACES = np.array(
    [
        [0, 1, 4, 3],
        [1, 2, 5, 4],
        [3, 4, 6, -1],
        [4, 5, 6, -1],
    ]
)
EDGE_NODES = np.array(
    [
        [0, 1],  # 0
        [0, 3],  # 1
        [1, 2],  # 2
        [1, 4],  # 3
        [2, 5],  # 4
        [3, 4],  # 5
        [3, 6],  # 6
        [4, 5],  # 7
        [4, 6],  # 8
        [5, 6],  # 9
    ]
)
EDGE_FACES = np.array(
    [
        [0, -1],
        [0, -1],
        [1, -1],
        [0, 1],
        [1, -1],
        [0, 2],
        [2, -1],
        [1, 3],
        [2, 3],
        [3, -1],
    ]
)
CENTROIDS = np.array(
    [
        [0.5, 0.5],
        [1.5, 0.5],
        [2.0 / 3.0, 4.0 / 3.0],
        [4.0 / 3.0, 4.0 / 3.0],
    ]
)
FFI = np.array([0, 0, 1, 1, 2, 2, 3, 3])
FFJ = np.array([1, 2, 0, 3, 0, 3, 1, 2])
FACE_FACE_CONNECTIVITY = sparse.coo_matrix((FFJ, (FFI, FFJ))).tocsr()
NFI = np.array([0, 1, 1, 2, 3, 3, 4, 4, 4, 4, 5, 5, 6, 6])
NFJ = np.array([0, 0, 1, 1, 0, 2, 0, 1, 2, 3, 1, 3, 2, 3])
NODE_FACE_CONNECTIVITY = sparse.coo_matrix((NFJ, (NFI, NFJ))).tocsr()


def grid2d(dataset=None, indexes=None, crs=None, attrs=None):
    grid = xugrid.Ugrid2d(
        node_x=VERTICES[:, 0],
        node_y=VERTICES[:, 1],
        fill_value=-1,
        face_node_connectivity=FACES,
        dataset=dataset,
        indexes=indexes,
        crs=crs,
        attrs=attrs,
    )
    return grid


def test_ugrid2d_init():
    grid = grid2d()
    assert grid.name == NAME
    assert grid._dataset is None
    assert grid.node_x.flags["C_CONTIGUOUS"]
    assert grid.node_y.flags["C_CONTIGUOUS"]
    assert grid._edge_node_connectivity is None
    assert grid._face_edge_connectivity is None


def test_safe_attrs():
    # .attrs should return a copy
    grid = grid2d()
    assert grid.attrs == grid.attrs
    assert grid._attrs is not grid.attrs


def test_ugrid2d_alternative_init():
    custom_attrs = {
        "node_dimension": "nNetNode",
        "name": "mesh1d",
        "node_coordinates": "mesh1d_node_x mesh1d_node_y",
    }
    indexes = {"node_x": "mesh1d_node_x", "node_y": "mesh1d_node_y"}
    grid = grid2d(attrs=custom_attrs, indexes=indexes)
    assert grid.node_dimension == "nNetNode"
    assert grid.name == NAME
    # name in attrs should be overwritten by given name.
    assert grid._attrs["name"] == NAME

    with pytest.raises(ValueError, match="Provide either dataset or attrs, not both"):
        grid2d(dataset=xr.Dataset, attrs=custom_attrs)

    with pytest.raises(ValueError, match="indexes must be provided for dataset"):
        grid2d(dataset=xr.Dataset, indexes=None)

    with pytest.raises(ValueError, match="indexes must be provided for attrs"):
        grid = grid2d(attrs=custom_attrs)


def test_ugrid2d_properties():
    grid = grid2d()
    assert grid.edge_dimension == f"{NAME}_nEdges"
    assert grid.node_dimension == f"{NAME}_nNodes"
    assert grid.face_dimension == f"{NAME}_nFaces"
    assert grid.n_node == 7
    assert grid.n_edge == 10
    assert grid.n_face == 4
    assert grid.n_max_node_per_face == 4
    assert grid.facets == {
        "node": grid.node_dimension,
        "edge": grid.edge_dimension,
        "face": grid.face_dimension,
    }
    assert np.array_equal(grid.n_node_per_face, [4, 4, 3, 3])
    assert np.allclose(grid.node_coordinates, VERTICES)
    assert grid.bounds == (0.0, 0.0, 2.0, 2.0)
    assert isinstance(grid.node_node_connectivity, sparse.csr_matrix)
    assert isinstance(grid.node_edge_connectivity, sparse.csr_matrix)
    assert isinstance(grid.directed_node_node_connectivity, sparse.csr_matrix)
    assert isinstance(grid.directed_edge_edge_connectivity, sparse.csr_matrix)
    edge_node_coords = grid.edge_node_coordinates
    face_node_coords = grid.face_node_coordinates
    assert edge_node_coords.shape == (10, 2, 2)
    assert face_node_coords.shape == (4, 4, 2)
    assert grid.edge_length.shape == (grid.n_edge,)
    assert grid.area.shape == (grid.n_face,)
    assert grid.perimeter.shape == (grid.n_face,)
    are_nan = np.isnan(face_node_coords)
    assert are_nan[2:, -1:, :].all()
    assert not are_nan[:, :-1, :].any()
    assert isinstance(grid.attrs, dict)
    coords = grid.coords
    assert isinstance(coords, dict)
    assert np.array_equal(coords[grid.node_dimension], grid.node_coordinates)
    assert np.array_equal(coords[grid.edge_dimension], grid.edge_coordinates)
    assert np.array_equal(coords[grid.face_dimension], grid.face_coordinates)

    with pytest.raises(ValueError, match="start_index must be 0 or 1, received: 2"):
        grid.start_index = 2
    grid.start_index = 1
    assert grid._start_index == 1

    assert isinstance(grid.node_kdtree, spatial.KDTree)
    assert isinstance(grid.edge_kdtree, spatial.KDTree)
    assert isinstance(grid.face_kdtree, spatial.KDTree)


def test_validate_edge_node_connectivity():
    # Full test at test_connectivity
    grid = grid2d()
    valid = grid.validate_edge_node_connectivity()
    assert isinstance(valid, np.ndarray)
    assert valid.size == grid.n_edge
    assert valid.all()


def test_ugrid2d_edge_bounds():
    grid = grid2d()
    expected = np.array(
        [
            [0.0, 0.0, 1.0, 0.0],
            [0.0, 0.0, 0.0, 1.0],
            [1.0, 0.0, 2.0, 0.0],
            [1.0, 0.0, 1.0, 1.0],
            [2.0, 0.0, 2.0, 1.0],
            [0.0, 1.0, 1.0, 1.0],
            [0.0, 1.0, 1.0, 2.0],
            [1.0, 1.0, 2.0, 1.0],
            [1.0, 1.0, 1.0, 2.0],
            [1.0, 1.0, 2.0, 2.0],
        ]
    )
    actual = grid.edge_bounds
    assert actual.shape == (10, 4)
    assert np.allclose(actual, expected)


def test_ugrid2d_face_bounds():
    grid = grid2d()
    expected = np.array(
        [
            [0.0, 0.0, 1.0, 1.0],
            [1.0, 0.0, 2.0, 1.0],
            [0.0, 1.0, 1.0, 2.0],
            [1.0, 1.0, 2.0, 2.0],
        ]
    )
    actual = grid.face_bounds
    assert actual.shape == (4, 4)
    assert np.allclose(actual, expected)


def test_set_crs():
    grid = grid2d()
    grid.set_crs("epsg:28992")
    assert grid.crs == pyproj.CRS.from_epsg(28992)


def test_ugrid2d_update_coordinate_attrs():
    grid = grid2d()
    obj = xr.DataArray(np.ones(grid.n_face), dims=(grid.face_dimension,))
    obj = grid.assign_face_coords(obj)
    grid._indexes["face_x"] = "mesh2d_face_x"
    grid._indexes["face_y"] = "mesh2d_face_y"
    grid.set_crs(epsg=4326)
    grid._update_coordinate_attrs(obj)
    assert obj["mesh2d_face_x"].attrs["standard_name"] == "longitude"
    assert obj["mesh2d_face_y"].attrs["standard_name"] == "latitude"


def test_ugrid2d_assign_derived_coordinates():
    grid = grid2d()
    obj = xr.DataArray(np.ones(grid.n_face), dims=(grid.face_dimension,))
    obj = grid._assign_derived_coords(obj)
    assert "mesh2d_face_x" in obj.coords
    assert "mesh2d_face_y" in obj.coords


def test_to_crs():
    grid = grid2d()
    grid.set_crs("epsg:4326")
    reprojected = grid.to_crs("epsg:28992")
    assert reprojected.crs == pyproj.CRS.from_epsg(28992)
    assert (~(grid.node_coordinates == reprojected.node_coordinates)).all()


def test_to_dataset():
    def check_attrs(ds):
        attrs = ds[NAME].attrs.copy()
        attrs.pop("cf_role")
        attrs.pop("long_name")
        attrs.pop("topology_dimension")
        ds_contents = tuple(ds.dims) + tuple(ds.coords) + tuple(ds.data_vars)
        for values in attrs.values():
            # e.g node_coordinates are joined by a whitespace.
            for value in values.split(" "):
                assert value in ds_contents

    grid = grid2d()
    ds = grid.to_dataset()
    assert isinstance(ds, xr.Dataset)
    assert f"{NAME}" in ds
    assert f"{NAME}_nNodes" in ds.dims
    assert f"{NAME}_nFaces" in ds.dims
    assert f"{NAME}_node_x" in ds.coords
    assert f"{NAME}_node_y" in ds.coords
    assert f"{NAME}_face_nodes" in ds
    check_attrs(ds)

    ds = grid.to_dataset(optional_attributes=True)
    assert f"{NAME}_edge_nodes" in ds
    assert f"{NAME}_face_nodes" in ds
    assert f"{NAME}_face_edges" in ds
    assert f"{NAME}_face_faces" in ds
    assert f"{NAME}_edge_faces" in ds
    assert f"{NAME}_boundary_nodes" in ds
    assert f"{NAME}_face_x" in ds
    assert f"{NAME}_face_y" in ds
    assert f"{NAME}_edge_x" in ds
    assert f"{NAME}_edge_y" in ds
    check_attrs(ds)


def test_find_ugrid_dim():
    grid = grid2d()
    da = xr.DataArray(data=np.ones((grid.n_face,)), dims=[grid.face_dimension])
    assert grid.find_ugrid_dim(da) == grid.face_dimension

    weird = xr.DataArray(
        data=np.ones((grid.n_face, grid.n_node)),
        dims=[grid.face_dimension, grid.node_dimension],
    )
    with pytest.raises(
        ValueError,
        match="UgridDataArray should contain exactly one of the UGRID dimension",
    ):
        grid.find_ugrid_dim(weird)


def test_ugrid2d_set_node_coords():
    grid = grid2d()
    ds = xr.Dataset()
    lonvalues = VERTICES[:, 0] + 10.0
    latvalues = VERTICES[:, 1] + 10.0
    ds["lon"] = xr.DataArray(lonvalues, dims=[grid.node_dimension])
    ds["lat"] = xr.DataArray(latvalues, dims=[grid.node_dimension])
    ds["lon with space"] = ds["lon"]
    ds["lat with space"] = ds["lat"]
    ds["short_lon"] = xr.DataArray(np.arange(6.0), dims=["short_node"])
    ds["long_lat"] = xr.DataArray(np.arange(8.0), dims=["long_node"])

    with pytest.raises(ValueError, match="coordinate names may not contain spaces"):
        grid.set_node_coords("lon with space", "lat with space", ds)
    with pytest.raises(
        ValueError, match="shape of node_x does not match n_node of grid: "
    ):
        grid.set_node_coords("short_lon", "lat", ds)
    with pytest.raises(
        ValueError, match="shape of node_y does not match n_node of grid: "
    ):
        grid.set_node_coords("lon", "long_lat", ds)

    grid.set_node_coords("lon", "lat", ds, is_projected=False)
    assert np.allclose(grid.node_x, lonvalues)
    assert np.allclose(grid.node_y, latvalues)
    assert grid._indexes["node_x"] == "lon"
    assert grid._indexes["node_y"] == "lat"
    assert not grid.is_projected


def test_ugrid2d_dataset_roundtrip():
    grid = grid2d()
    ds = grid.to_dataset()
    grid2 = xugrid.Ugrid2d.from_dataset(ds)
    assert isinstance(grid2._dataset, xr.Dataset)
    assert grid2._dataset == ds


def test_ugrid2d_dataset_no_mutation():
    grid = grid2d()
    ds = grid.to_dataset()
    # Test a non-default fill value
    face_nodes = ds["mesh2d_face_nodes"]
    face_nodes = face_nodes.where(face_nodes != -1, other=-999)
    face_nodes.attrs["_FillValue"] = -999
    ds["mesh2d_face_nodes"] = face_nodes
    reference = ds.copy(deep=True)
    xugrid.Ugrid2d.from_dataset(ds)
    assert ds.identical(reference)


@pytest.mark.parametrize("edge_start_index", [0, 1])
@pytest.mark.parametrize("face_start_index", [0, 1])
def test_ugrid2d_from_dataset__different_start_index(
    face_start_index, edge_start_index
):
    grid = grid2d()
    ds = grid.to_dataset(optional_attributes=True)  # include edge_nodes
    faces = ds["mesh2d_face_nodes"].to_numpy()
    faces[faces != -1] += face_start_index
    ds["mesh2d_face_nodes"].attrs["start_index"] = face_start_index
    ds["mesh2d_edge_nodes"] += edge_start_index
    ds["mesh2d_edge_nodes"].attrs["start_index"] = edge_start_index
    new = xugrid.Ugrid2d.from_dataset(ds)
    assert new.start_index == face_start_index
    assert np.array_equal(new.face_node_connectivity, grid.face_node_connectivity)
    assert np.array_equal(new.edge_node_connectivity, grid.edge_node_connectivity)


def test_ugrid2d_from_meshkernel():
    # Setup a meshkernel Mesh2d mimick
    class Mesh2d(NamedTuple):
        node_x: np.ndarray
        node_y: np.ndarray
        face_nodes: np.ndarray
        nodes_per_face: np.ndarray
        edge_nodes: np.ndarray

    mesh2d = Mesh2d(
        node_x=np.array([0.0, 1.0, 2.0, 3.0, 0.0, 1.0, 2.0, 3.0, 0.0, 1.0, 2.0, 3.0]),
        node_y=np.array([0.0, 0.0, 0.0, 0.0, 1.0, 1.0, 1.0, 1.0, 2.0, 2.0, 2.0, 2.0]),
        face_nodes=np.array(
            [0, 1, 5, 4, 1, 2, 6, 5, 2, 3, 7, 6, 4, 5, 9, 8, 5, 6, 10, 9, 6, 7, 11, 10]
        ),
        nodes_per_face=np.array([4, 4, 4, 4, 4, 4]),
        edge_nodes=np.array(
            [
                4,
                8,
                5,
                6,
                5,
                9,
                6,
                7,
                6,
                10,
                7,
                11,
                8,
                9,
                9,
                10,
                10,
                11,
                0,
                1,
                0,
                4,
                1,
                2,
                1,
                5,
                2,
                3,
                2,
                6,
                3,
                7,
                4,
                5,
            ]
        ),
    )

    grid = xugrid.Ugrid2d.from_meshkernel(mesh2d)
    grid.plot()
    assert grid.n_face == 6
    assert np.allclose(mesh2d.node_x, grid.node_x)
    assert np.allclose(mesh2d.node_y, grid.node_y)
    assert np.allclose(grid.face_node_connectivity, mesh2d.face_nodes.reshape((6, 4)))
    assert np.allclose(grid.edge_node_connectivity, mesh2d.edge_nodes.reshape((-1, 2)))


def test_assign_node_coords():
    grid = grid2d()
    ds = xr.Dataset()
    # Place some data on the grid facets.
    ds["a"] = xr.DataArray([1, 2, 3, 4, 5, 6, 7], dims=[f"{NAME}_nNodes"])
    with_coords = grid.assign_node_coords(ds)
    assert f"{NAME}_node_x" in with_coords
    assert f"{NAME}_node_y" in with_coords
    assert np.array_equal(with_coords[f"{NAME}_node_x"], grid.node_x)
    assert np.array_equal(with_coords[f"{NAME}_node_y"], grid.node_y)


def test_assign_edge_coords():
    grid = grid2d()
    ds = xr.Dataset()
    # Place some data on the grid facets.
    ds["a"] = xr.DataArray([1, 2, 3, 4, 5, 6, 7], dims=[f"{NAME}_nNodes"])
    with_coords = grid.assign_edge_coords(ds)
    assert f"{NAME}_edge_x" in with_coords
    assert f"{NAME}_edge_y" in with_coords
    assert np.array_equal(with_coords[f"{NAME}_edge_x"], grid.edge_x)
    assert np.array_equal(with_coords[f"{NAME}_edge_y"], grid.edge_y)


def test_assign_face_coords():
    grid = grid2d()
    ds = xr.Dataset()
    # Place some data on the grid facets.
    ds["a"] = xr.DataArray([1, 2, 3, 4, 5, 6, 7], dims=[f"{NAME}_nNodes"])
    with_coords = grid.assign_face_coords(ds)
    assert f"{NAME}_face_x" in with_coords
    assert f"{NAME}_face_y" in with_coords
    assert np.array_equal(with_coords[f"{NAME}_face_x"], grid.face_x)
    assert np.array_equal(with_coords[f"{NAME}_face_y"], grid.face_y)


def test_clear_geometry_properties():
    grid = grid2d()
    for attr in [
        "_mesh",
        "_meshkernel",
        "_celltree",
        "_node_kdtree",
        "_edge_kdtree",
        "_face_kdtree",
        "_centroids",
        "_xmin",
        "_xmax",
        "_ymin",
        "_ymax",
        "_edge_x",
        "_edge_y",
        "_triangulation",
        "_voronoi_topology",
        "_centroid_triangulation",
    ]:
        setattr(grid, attr, 1)
        grid._clear_geometry_properties()
        assert getattr(grid, attr) is None


def test_topology_dimension():
    grid = grid2d()
    assert grid.topology_dimension == 2


def test_dimensions():
    grid = grid2d()
    assert grid.node_dimension == f"{NAME}_nNodes"
    assert grid.edge_dimension == f"{NAME}_nEdges"
    assert grid.face_dimension == f"{NAME}_nFaces"
    assert grid.dims == {
        f"{NAME}_nNodes",
        f"{NAME}_nEdges",
        f"{NAME}_nFaces",
    }
    assert grid.sizes == {
        f"{NAME}_nNodes": 7,
        f"{NAME}_nEdges": 10,
        f"{NAME}_nFaces": 4,
    }
    with pytest.warns(FutureWarning):
        assert grid.dimensions == grid.sizes


def test_edge_node_connectivity():
    grid = grid2d()
    edge_nodes = grid.edge_node_connectivity
    assert grid._edge_node_connectivity is not None
    assert grid._face_edge_connectivity is not None
    assert np.allclose(edge_nodes, EDGE_NODES)


def test_edge_face_connectivity():
    grid = grid2d()
    edge_faces = grid.edge_face_connectivity
    assert grid._edge_node_connectivity is not None
    assert grid._face_edge_connectivity is not None
    assert np.allclose(edge_faces, EDGE_FACES)


def test_centroids():
    grid = grid2d()
    assert np.allclose(grid.centroids, CENTROIDS)
    assert np.allclose(grid.face_coordinates, CENTROIDS)
    assert np.allclose(grid.face_x, CENTROIDS[:, 0])
    assert np.allclose(grid.face_y, CENTROIDS[:, 1])


def test_face_face_connectivity():
    grid = grid2d()
    face_face = grid.face_face_connectivity
    assert isinstance(face_face, sparse.csr_matrix)
    assert np.array_equal(face_face.indptr, FACE_FACE_CONNECTIVITY.indptr)
    assert np.array_equal(face_face.indices, FACE_FACE_CONNECTIVITY.indices)


def test_node_face_connectivity():
    grid = grid2d()
    node_face = grid.node_face_connectivity
    assert isinstance(node_face, sparse.csr_matrix)
    assert np.array_equal(node_face.indptr, NODE_FACE_CONNECTIVITY.indptr)
    assert np.array_equal(node_face.indices, NODE_FACE_CONNECTIVITY.indices)


def test_connectivity_matrix():
    grid = grid2d()
    with pytest.raises(
        ValueError, match="Expected mesh2d_nNodes or mesh2d_nFaces; got: mesh2d_nEdges"
    ):
        grid.get_connectivity_matrix(dim=grid.edge_dimension, xy_weights=False)

    connectivity = grid.get_connectivity_matrix(grid.face_dimension, xy_weights=True)
    assert isinstance(connectivity, sparse.csr_matrix)
    assert np.array_equal(connectivity.indices, [1, 2, 0, 3, 0, 3, 1, 2])

    connectivity = grid.get_connectivity_matrix(grid.node_dimension, xy_weights=True)
    assert isinstance(connectivity, sparse.csr_matrix)
    assert np.array_equal(
        connectivity.indices,
        [1, 3, 0, 2, 4, 1, 5, 0, 4, 6, 1, 3, 5, 6, 2, 4, 6, 3, 4, 5],
    )


def test_get_coordinates():
    grid = grid2d()
    with pytest.raises(
        ValueError,
        match="Expected mesh2d_nNodes, mesh2d_nEdges, or mesh2d_nFaces; got: abc",
    ):
        grid.get_coordinates(dim="abc")

    assert isinstance(grid.get_coordinates(grid.node_dimension), np.ndarray)
    assert isinstance(grid.get_coordinates(grid.edge_dimension), np.ndarray)
    assert isinstance(grid.get_coordinates(grid.face_dimension), np.ndarray)


def test_voronoi_topology():
    grid = grid2d()
    vertices, faces, face_index = grid.voronoi_topology
    expected_exterior = np.array(
        [
            [0.5, 0.0],
            [0.0, 0.5],
            [1.5, 0.0],
            [2.0, 0.5],
            [0.5, 1.5],
            [1.5, 1.5],
        ]
    )
    expected_vertices = np.vstack([CENTROIDS, expected_exterior])
    assert np.allclose(vertices, expected_vertices)
    assert isinstance(faces, np.ndarray)
    expected_faces = np.array(
        [
            [0, 1, 3, 2],
            [4, 0, 5, -1],
            [4, 6, 1, 0],
            [6, 7, 1, -1],
            [5, 0, 2, 8],
            [1, 7, 9, 3],
            [2, 3, 9, 8],
        ]
    )
    assert np.array_equal(faces, expected_faces)
    assert np.array_equal(face_index, [0, 1, 2, 3, 0, 0, 1, 1, 2, 3])


def test_centroid_triangulation():
    grid = grid2d()
    (x, y, triangles), face_index = grid.centroid_triangulation
    assert np.allclose(x, list(CENTROIDS[:, 0]) + [0.5, 0.0, 1.5, 2.0, 0.5, 1.5])
    assert np.allclose(y, list(CENTROIDS[:, 1]) + [0.0, 0.5, 0.0, 0.5, 1.5, 1.5])
    expected_triangles = np.array(
        [
            [0, 1, 3],
            [0, 3, 2],
            [4, 0, 5],
            [4, 6, 1],
            [4, 1, 0],
            [6, 7, 1],
            [5, 0, 2],
            [5, 2, 8],
            [1, 7, 9],
            [1, 9, 3],
            [2, 3, 9],
            [2, 9, 8],
        ]
    )
    assert np.array_equal(triangles, expected_triangles)
    assert np.array_equal(face_index, [0, 1, 2, 3, 0, 0, 1, 1, 2, 3])


def test_triangulation():
    grid = grid2d()
    (x, y, triangles), face_index = grid.triangulation
    expected_triangles = np.array(
        [
            [0, 1, 4],
            [0, 4, 3],
            [1, 2, 5],
            [1, 5, 4],
            [3, 4, 6],
            [4, 5, 6],
        ]
    )
    assert np.allclose(x, VERTICES[:, 0])
    assert np.allclose(y, VERTICES[:, 1])
    assert np.array_equal(triangles, expected_triangles)
    assert np.array_equal(face_index, [0, 0, 1, 1, 2, 3])


def test_exterior_edges():
    grid = grid2d()
    assert np.array_equal(grid.exterior_edges, [0, 1, 2, 4, 6, 9])


def test_exterior_faces():
    grid = grid2d()
    assert np.array_equal(grid.exterior_faces, [0, 1, 2, 3])


def test_celltree():
    grid = grid2d()
    tree = grid.celltree
    assert isinstance(tree, numba_celltree.CellTree2d)


def test_locate_points():
    grid = grid2d()
    assert np.array_equal(grid.locate_points(CENTROIDS), [0, 1, 2, 3])
    # Test tolerance
    centroids_offset = [[-0.01, 1.0], [-0.01, 0.5]]
    assert np.array_equal(grid.locate_points(centroids_offset, 0.011), [0, 0])


def test_locate_bounding_box():
    grid = grid2d()
    faces = grid.locate_bounding_box(1.25, 0.25, 2.5, 1.5)
    assert np.allclose(faces, [1, 3])


def test_compute_barycentric_weights():
    grid = grid2d()
    xy = np.array(
        [
            [0.0, 0.0],
            [0.5, 0.5],
            [1.5, 0.5],
            [0.5, 1.5],
            [2.0, 2.0],
        ]
    )
    expected_face = np.array([0, 0, 1, 2, -1])
    expected_weights = np.array(
        [
            [1.0, 0.0, 0.0, 0.0],
            [0.25, 0.25, 0.25, 0.25],
            [0.25, 0.25, 0.25, 0.25],
            [0.5, 0.0, 0.5, 0.0],
            [0.0, 0.0, 0.0, 0.0],
        ]
    )
    face, weights = grid.compute_barycentric_weights(xy)
    assert np.array_equal(face, expected_face)
    assert np.allclose(weights, expected_weights)
    # Test with tolerance. First point goes out of bounds, tolerance shouldn't
    # matter.
    xy[:, 0] -= 0.01
    face, weights = grid.compute_barycentric_weights(xy, tolerance=0.01)
    expected_face = np.array([-1, 0, 1, 2, -1])
    expected_weights = np.array(
        [
            [0.0, 0.0, 0.0, 0.0],
            [0.25, 0.25, 0.25, 0.25],
            [0.25, 0.25, 0.25, 0.25],
            [0.5, 0.0, 0.5, 0.0],
            [0.0, 0.0, 0.0, 0.0],
        ]
    )
    assert np.array_equal(face, expected_face)
    assert np.allclose(weights, expected_weights, atol=0.05)


def test_rasterize():
    grid = grid2d()
    x, y, index = grid.rasterize(resolution=0.5)
    expected_index = np.array(
        [
            [-1, 2, 3, -1],
            [2, 2, 3, 3],
            [0, 0, 1, 1],
            [0, 0, 1, 1],
        ]
    )
    assert np.allclose(x, [0.25, 0.75, 1.25, 1.75])
    assert np.allclose(y, [1.75, 1.25, 0.75, 0.25])
    assert np.array_equal(index, expected_index)

    # Test with alternative bounds
    bounds = (-1.0, -1.0, 2.0, 2.0)
    x, y, index = grid.rasterize(resolution=0.5, bounds=bounds)
    expected_index = np.array(
        [
            [-1, -1, -1, 2, 3, -1],
            [-1, -1, 2, 2, 3, 3],
            [-1, -1, 0, 0, 1, 1],
            [-1, -1, 0, 0, 1, 1],
            [-1, -1, -1, -1, -1, -1],
            [-1, -1, -1, -1, -1, -1],
        ]
    )
    assert np.allclose(x, [-0.75, -0.25, 0.25, 0.75, 1.25, 1.75])
    assert np.allclose(y, [1.75, 1.25, 0.75, 0.25, -0.25, -0.75])
    assert np.array_equal(index, expected_index)


class TestUgrid2dSelection:
    @pytest.fixture(autouse=True)
    def setup(self):
        self.grid = grid2d()
        self.obj = xr.DataArray([0, 1, 2, 3], dims=[self.grid.face_dimension])

    def test_sel_points(self):
        x = [0.5, 1.5]
        y = [0.5, 1.25]

        with pytest.raises(ValueError, match="method must be one of"):
            self.grid.sel_points(obj=self.obj, x=x, y=y, method="nothing")
        with pytest.raises(ValueError, match="out_of_bounds must be one of"):
            self.grid.sel_points(obj=self.obj, x=x, y=y, out_of_bounds="nothing")
        with pytest.raises(ValueError, match="shape of x does not match shape of y"):
            self.grid.sel_points(obj=self.obj, x=[0.5, 1.5], y=[0.5])
        with pytest.raises(ValueError, match="x and y must be 1d"):
            self.grid.sel_points(obj=self.obj, x=[x], y=[y])

        actual = self.grid.sel_points(obj=self.obj, x=x, y=y)
        assert isinstance(actual, xr.DataArray)

        dim = f"{NAME}_points"
        expected = xr.DataArray(
            data=[0, 3],
            coords={
                f"{NAME}_x": (dim, x),
                f"{NAME}_y": (dim, y),
            },
            dims=[dim],
        )
        assert expected.equals(actual)

    def test_sel_points_out_of_bounds(self):
        x = [-10.0, 0.5, -20.0, 1.5, -30.0]
        y = [-10.0, 0.5, -20.0, 1.25, -30.0]

        with pytest.raises(
            ValueError, match="Not all points are located on the topology"
        ):
            self.grid.sel_points(obj=self.obj, x=x, y=y, out_of_bounds="raise")

        actual = self.grid.sel_points(obj=self.obj, x=x, y=y, out_of_bounds="drop")

        with pytest.warns(
            UserWarning, match="Not all points are located on the topology"
        ):
            actual = self.grid.sel_points(obj=self.obj, x=x, y=y, out_of_bounds="warn")
            assert np.allclose(actual, [np.nan, 0, np.nan, 3, np.nan], equal_nan=True)

        actual = self.grid.sel_points(obj=self.obj, x=x, y=y, out_of_bounds="ignore")
        assert np.allclose(actual, [np.nan, 0, np.nan, 3, np.nan], equal_nan=True)

        actual = self.grid.sel_points(
            obj=self.obj, x=x, y=y, out_of_bounds="ignore", fill_value=-1
        )
        assert np.allclose(actual, [-1, 0, -1, 3, -1])
        # Case with tolerance, tolerance shouldn't affect results since points
        # are out of bounds
        actual = self.grid.sel_points(
            obj=self.obj, x=x, y=y, out_of_bounds="drop", tolerance=11.0
        )

    def test_sel_points_multiple_dims(self):
        grid = self.grid
        ds = xr.Dataset(
            {
                "face_data": self.obj,
                "node_data": xr.DataArray(
                    np.arange(grid.n_node), dims=(grid.node_dimension)
                ),
                "edge_data": xr.DataArray(
                    np.arange(grid.n_edge), dims=(grid.edge_dimension)
                ),
            }
        )
        x = [-10.0, 0.5, -20.0, 1.5, -30.0]
        y = [-10.0, 0.5, -20.0, 1.25, -30.0]

        # Test with out of bounds ignored
        face_actual = grid.sel_points(
            obj=ds["face_data"], x=x, y=y, out_of_bounds="ignore"
        )
        np.testing.assert_allclose(
            face_actual, [np.nan, 0, np.nan, 3, np.nan], equal_nan=True
        )
        node_actual = grid.sel_points(
            obj=ds["node_data"], x=x, y=y, out_of_bounds="ignore"
        )
        np.testing.assert_allclose(
            node_actual, [np.nan, 0, np.nan, 4, np.nan], equal_nan=True
        )
        edge_actual = grid.sel_points(
            obj=ds["node_data"], x=x, y=y, out_of_bounds="ignore"
        )
        np.testing.assert_allclose(
            edge_actual, [np.nan, 0, np.nan, 4, np.nan], equal_nan=True
        )

        # With drop
        face_actual = grid.sel_points(
            obj=ds["face_data"], x=x, y=y, out_of_bounds="drop"
        )
        np.testing.assert_allclose(face_actual, [0, 3], equal_nan=True)
        node_actual = grid.sel_points(
            obj=ds["node_data"], x=x, y=y, out_of_bounds="drop"
        )
        np.testing.assert_allclose(node_actual, [0, 4], equal_nan=True)
        edge_actual = grid.sel_points(
            obj=ds["node_data"], x=x, y=y, out_of_bounds="drop"
        )
        np.testing.assert_allclose(edge_actual, [0, 4], equal_nan=True)

        with pytest.warns(
            UserWarning, match="Not all points are located on the topology"
        ):
            node_actual = grid.sel_points(
                obj=ds["node_data"], x=x, y=y, out_of_bounds="warn"
            )
            np.testing.assert_allclose(
                node_actual, [np.nan, 0, np.nan, 4, np.nan], equal_nan=True
            )

        with pytest.raises(
            ValueError, match="Not all points are located on the topology"
        ):
            grid.sel_points(obj=ds["node_data"], x=x, y=y, out_of_bounds="raise")

        # Test with method "nearest"
        # Core dimension should change (face), but secondary dimension should not: they are always nearest.
        x = [0.4]
        y = [0.99]
        face_actual = grid.sel_points(
            obj=ds["face_data"], x=x, y=y, out_of_bounds="ignore"
        )
        np.testing.assert_allclose(face_actual, [0], equal_nan=True)
        node_actual = grid.sel_points(
            obj=ds["node_data"], x=x, y=y, out_of_bounds="ignore"
        )
        np.testing.assert_allclose(node_actual, [3], equal_nan=True)
        edge_actual = grid.sel_points(
            obj=ds["edge_data"], x=x, y=y, out_of_bounds="ignore"
        )
        np.testing.assert_allclose(edge_actual, [5], equal_nan=True)

        face_actual = grid.sel_points(
            obj=ds["face_data"], x=x, y=y, method="nearest", out_of_bounds="ignore"
        )
        np.testing.assert_allclose(face_actual, [2], equal_nan=True)
        node_actual = grid.sel_points(
            obj=ds["node_data"], x=x, y=y, method="nearest", out_of_bounds="ignore"
        )
        np.testing.assert_allclose(node_actual, [3], equal_nan=True)
        face_actual = grid.sel_points(
            obj=ds["edge_data"], x=x, y=y, method="nearest", out_of_bounds="ignore"
        )
        np.testing.assert_allclose(face_actual, [5], equal_nan=True)

    def test_validate_indexer(self):
        with pytest.raises(ValueError, match="slice stop should be larger than"):
            self.grid._validate_indexer(slice(2, 0))
        with pytest.raises(ValueError, match="step should be None"):
            self.grid._validate_indexer(slice(None, 2, 1))
        with pytest.raises(ValueError, match="step should be None"):
            self.grid._validate_indexer(slice(0, None, 1))

        expected = np.arange(0.0, 2.0, 0.5)
        assert np.allclose(self.grid._validate_indexer(slice(0, 2, 0.5)), expected)
        assert self.grid._validate_indexer(slice(None, 2)) == slice(None, 2)
        assert self.grid._validate_indexer(slice(0, None)) == slice(0, None)

        with pytest.raises(TypeError, match="Invalid indexer type"):
            self.grid._validate_indexer((0, 1, 2))

        # list
        actual = self.grid._validate_indexer([0.0, 1.0, 2.0])
        assert isinstance(actual, np.ndarray)
        assert np.allclose(actual, [0.0, 1.0, 2.0])

        # numpy array
        actual = self.grid._validate_indexer(np.array([0.0, 1.0, 2.0]))
        assert isinstance(actual, np.ndarray)
        assert np.allclose(actual, [0.0, 1.0, 2.0])

        # xarray DataArray
        indexer = xr.DataArray([0.0, 1.0, 2.0], {"x": [0, 1, 2]}, ["x"])
        actual = self.grid._validate_indexer(indexer)
        assert isinstance(actual, np.ndarray)
        assert np.allclose(actual, [0.0, 1.0, 2.0])

        # float
        actual = self.grid._validate_indexer(1.0)
        assert isinstance(actual, np.ndarray)
        assert np.allclose(actual, [1.0])

        # int
        actual = self.grid._validate_indexer(1)
        assert isinstance(actual, np.ndarray)
        assert np.allclose(actual, [1])

    def test_sel__bounding_box(self):
        def check_output(actual, expected):
            assert isinstance(actual, tuple)
            new_obj, new_grid = actual
            assert isinstance(new_obj, xr.DataArray)
            assert isinstance(new_grid, xugrid.Ugrid2d)
            assert new_obj.dims[0] == f"{NAME}_nFaces"
            assert new_grid.face_dimension == f"{NAME}_nFaces"
            assert np.array_equal(new_obj.values, expected)

        actual = self.grid.sel(obj=self.obj, x=slice(0.0, 2.0), y=slice(0.0, 1.0))
        check_output(actual, [0, 1])

        actual = self.grid.sel(obj=self.obj, x=slice(None, None), y=slice(None, 1.0))
        check_output(actual, [0, 1])

        actual = self.grid.sel(obj=self.obj, x=slice(0.0, 1.0), y=slice(0.0, 2.0))
        check_output(actual, [0, 2])

        actual = self.grid.sel(obj=self.obj, x=slice(None, 1.0), y=slice(None, None))
        check_output(actual, [0, 2])

        for x, y in zip([None, None, slice(0, 2)], [None, slice(0, 2), None]):
            actual = self.grid.sel(obj=self.obj, x=x, y=y)
            check_output(actual, [0, 1, 2, 3])

        # Check default arguments, should return entire grid
        actual = self.grid.sel(obj=self.obj)
        check_output(actual, [0, 1, 2, 3])

    def test_sel__points_from_scalar(self):
        def check_output(actual):
            assert isinstance(actual, xr.DataArray)
            dim = f"{NAME}_points"
            expected = xr.DataArray(
                data=[0],
                coords={
                    f"{NAME}_x": (dim, [0.5]),
                    f"{NAME}_y": (dim, [0.5]),
                },
                dims=[dim],
            )
            assert expected.equals(actual)

        actual = self.grid.sel(obj=self.obj, x=0.5, y=0.5)
        check_output(actual)

        actual = self.grid.sel(obj=self.obj, x=[0.5], y=[0.5])
        check_output(actual)

        with pytest.raises(TypeError, match="Invalid indexer type"):
            self.grid.sel(obj=self.obj, x=(0.5,), y=[0.5])

    def test_sel__points_from_arrays_and_slice(self):
        def check_output(actual):
            assert isinstance(actual, xr.DataArray)
            dim = f"{NAME}_points"
            expected = xr.DataArray(
                data=[0, 0, 1, 2, 2, 3],
                coords={
                    f"{NAME}_x": (dim, [0.4, 0.8, 1.2, 0.4, 0.8, 1.2]),
                    f"{NAME}_y": (dim, [0.5, 0.5, 0.5, 1.1, 1.1, 1.1]),
                },
                dims=[dim],
            )
            # This fails for some reason:
            # assert expected.equals(actual)
            assert np.array_equal(expected.values, actual.values)
            assert expected.dims == actual.dims
            x = f"{NAME}_x"
            y = f"{NAME}_y"
            assert np.allclose(expected[y].values, actual[y].values)
            assert np.allclose(expected[x].values, actual[x].values)

        x = [0.4, 0.8, 1.2]
        y = [0.5, 1.1]
        actual = self.grid.sel(obj=self.obj, x=x, y=y)
        check_output(actual)

        x = slice(0.4, 1.5, 0.4)  # Evaluates to: [0.4, 0.8, 1.2]
        actual = self.grid.sel(obj=self.obj, x=x, y=y)
        check_output(actual)

    def test_sel__edges_from_slice(self):
        with pytest.raises(ValueError, match="If x is a slice without steps"):
            self.grid.sel(obj=self.obj, x=slice(None, None), y=[0.25, 0.75])
        with pytest.raises(ValueError, match="If x is a slice without steps"):
            self.grid.sel(obj=self.obj, x=slice(None, None), y=slice(0.25, 1.0, 0.25))
        with pytest.raises(ValueError, match="If y is a slice without steps"):
            self.grid.sel(obj=self.obj, x=[0.25, 0.75], y=slice(None, None))

        actual = self.grid.sel(obj=self.obj, x=slice(None, None), y=0.5)
        assert isinstance(actual, xr.DataArray)
        dim = f"{NAME}_nFaces"
        expected = xr.DataArray(
            data=[0, 1],
            coords={
                f"{NAME}_x": (dim, [0.5, 1.5]),
                f"{NAME}_y": (dim, [0.5, 0.5]),
                f"{NAME}_s": (dim, [0.5, 1.5]),
            },
            dims=[dim],
        )
        assert expected.equals(actual)

        actual = self.grid.sel(obj=self.obj, x=0.5, y=slice(None, None))
        assert isinstance(actual, xr.DataArray)
        expected = xr.DataArray(
            data=[0, 2],
            coords={
                f"{NAME}_x": (dim, [0.5, 0.5]),
                f"{NAME}_y": (dim, [0.5, 1.25]),
                f"{NAME}_s": (dim, [0.5, 1.25]),
            },
            dims=[dim],
        )
        assert expected.equals(actual)

    def test_intersect_line_error(self):
        with pytest.raises(ValueError, match="Start and end coordinate pairs"):
            self.grid.intersect_line(
                obj=None, start=(0.0, 0.0, 0.0), end=(1.0, 1.0, 1.0)
            )

    def test_intersect_line(self):
        grid = self.grid
        obj = xr.DataArray([0, 1, 2, 3], dims=[grid.face_dimension])

        p0 = (0.0, 0.0)
        p1 = (2.0, 2.0)
        actual = grid.intersect_line(obj, start=p0, end=p1)
        sqrt2 = np.sqrt(2.0)
        assert isinstance(actual, xr.DataArray)
        assert actual.dims == (grid.face_dimension,)
        assert np.array_equal(actual.to_numpy(), [0, 3])
        assert np.allclose(actual[f"{NAME}_x"], [0.5, 1.25])
        assert np.allclose(actual[f"{NAME}_y"], [0.5, 1.25])
        assert np.allclose(actual[f"{NAME}_s"], [0.5 * sqrt2, 1.25 * sqrt2])

        actual = grid.intersect_line(obj, start=p1, end=p0)
        assert np.array_equal(actual.to_numpy(), [3, 0])

    def test_intersect_linestring(self):
        grid = self.grid
        obj = xr.DataArray([0, 1, 2, 3], dims=[grid.face_dimension])
        linestring = shapely.geometry.LineString(
            [
                [0.5, 0.5],
                [1.5, 0.5],
                [1.5, 1.5],
            ]
        )
        actual = grid.intersect_linestring(obj, linestring)
        assert isinstance(actual, xr.DataArray)
        assert actual.dims == (grid.face_dimension,)
        assert np.array_equal(actual.to_numpy(), [0, 1, 1, 3])
        assert np.allclose(actual[f"{NAME}_x"], [0.75, 1.25, 1.5, 1.5])
        assert np.allclose(actual[f"{NAME}_y"], [0.5, 0.5, 0.75, 1.25])
        assert np.allclose(actual[f"{NAME}_s"], [0.25, 0.75, 1.25, 1.75])


def test_topology_subset():
    grid = grid2d()
    face_index = np.array([1])
    actual = grid.topology_subset(face_index)
    assert np.array_equal(actual.face_node_connectivity, [[0, 1, 3, 2]])
    assert np.array_equal(actual.node_x, [1.0, 2.0, 1.0, 2.0])
    assert np.array_equal(actual.node_y, [0.0, 0.0, 1.0, 1.0])

    face_index = np.array([False, True, False, False])
    actual = grid.topology_subset(face_index)
    assert np.array_equal(actual.face_node_connectivity, [[0, 1, 3, 2]])
    assert np.array_equal(actual.node_x, [1.0, 2.0, 1.0, 2.0])
    assert np.array_equal(actual.node_y, [0.0, 0.0, 1.0, 1.0])

    # Entire mesh
    face_index = np.array([0, 1, 2, 3])
    actual = grid.topology_subset(face_index)
    assert actual is grid

    # Reordering
    face_index = np.array([3, 2, 1, 0])
    actual = grid.topology_subset(face_index)
    assert np.array_equal(
        actual.face_node_connectivity,
        grid.face_node_connectivity[::-1],
    )

    # Check that alternative attrs are preserved.
    grid = grid2d(
        attrs={"node_dimension": "nNetNode"},
        indexes={"node_x": "mesh2d_node_x", "node_y": "mesh2d_node_y"},
    )
    face_index = np.array([1])
    actual = grid.topology_subset(face_index)
    assert actual.node_dimension == "nNetNode"


def test_reindex_like():
    grid = grid2d()
    # Change face and edge_index.
    index = np.arange(grid.n_face)
    rev_index = index[::-1]
    edge_index = np.arange(grid.n_edge)
    rev_edge_index = edge_index[::-1]
    reordered = grid.topology_subset(rev_index)
    reordered._edge_node_connectivity = reordered._edge_node_connectivity[
        rev_edge_index
    ]

    face_da = xr.DataArray(rev_index, dims=(reordered.face_dimension,))
    edge_da = xr.DataArray(rev_edge_index, dims=(reordered.edge_dimension,))
    ds = xr.Dataset({"edge": edge_da, "face": face_da})
    reindexed = reordered.reindex_like(grid, obj=ds)
    assert np.array_equal(reindexed["face"], index)
    assert np.array_equal(reindexed["edge"], edge_index)


def test_triangulate():
    grid = grid2d()
    actual = grid.triangulate()
    assert isinstance(actual, xugrid.Ugrid2d)
    assert actual.n_face == 6


def test_tesselate_centroidal_voronoi():
    grid = grid2d()

    voronoi = grid.tesselate_centroidal_voronoi(add_exterior=False)
    assert isinstance(voronoi, xugrid.Ugrid2d)
    assert voronoi.n_face == 1

    voronoi = grid.tesselate_centroidal_voronoi(add_vertices=False)
    assert voronoi.n_face == 7

    voronoi = grid.tesselate_centroidal_voronoi()
    assert voronoi.n_face == 7

    faces = FACES.copy()
    faces[faces == -1] = -999
    grid = xugrid.Ugrid2d(
        node_x=VERTICES[:, 0],
        node_y=VERTICES[:, 1],
        fill_value=-999,
        face_node_connectivity=faces,
    )
    voronoi = grid.tesselate_centroidal_voronoi(add_exterior=True)
    vfaces = voronoi.face_node_connectivity
    fill_nodes = vfaces[vfaces < 0]
    assert (fill_nodes == -1).all()


def test_tesselate_circumcenter_voronoi():
    grid = grid2d()

    # Can only deal with triangular grids
    with pytest.raises(NotImplementedError):
        grid.tesselate_circumcenter_voronoi()

    # Now test with triangular grid
    vertices = np.array(
        [
            [0.0, 0.0],  # 0
            [2.0, 0.0],  # 1
            [1.0, 1.0],  # 2
            [2.0, 2.0],  # 3
            [0.0, 2.0],  # 4
        ]
    )
    faces = np.array(
        [
            [0, 1, 2],
            [1, 3, 2],
            [3, 4, 2],
            [4, 0, 2],
        ]
    )
    grid = xugrid.Ugrid2d(
        node_x=vertices[:, 0],
        node_y=vertices[:, 1],
        fill_value=-1,
        face_node_connectivity=faces,
    )
    voronoi = grid.tesselate_circumcenter_voronoi()
    assert voronoi.n_face == 5


def test_reverse_cuthill_mckee():
    grid = grid2d()
    new, index = grid.reverse_cuthill_mckee()
    assert isinstance(new, xugrid.Ugrid2d)
    assert np.allclose(new.node_coordinates, grid.node_coordinates)
    assert np.array_equal(index, [3, 2, 1, 0])


@requires_meshkernel
def test_mesh():
    grid = grid2d()
    assert isinstance(grid.mesh, mk.Mesh2d)


@requires_meshkernel
def test_meshkernel():
    grid = grid2d()
    assert isinstance(grid.meshkernel, mk.MeshKernel)


@pytest.mark.parametrize("xflip", [False, True])
@pytest.mark.parametrize("yflip", [False, True])
def test_from_structured_intervals1d(xflip: bool, yflip: bool):
    x = y = np.array([0.0, 1.0, 2.0])
    if xflip:
        x = np.flip(x)
    if yflip:
        y = np.flip(y)

    grid = xugrid.Ugrid2d.from_structured_intervals1d(x_intervals=x, y_intervals=y)
    assert isinstance(grid, xugrid.Ugrid2d)
    assert grid.n_face == 4

    # Make sure the orientation is still ccw if x is decreasing, y is increasing.
    dxy = np.diff(grid.face_node_coordinates, axis=1)
    clockwise = (
        xugrid.ugrid.connectivity.cross2d(dxy[:, :-1], dxy[:, 1:]).sum(axis=1) < 0
    )
    assert not clockwise.any()


def test_from_structured_intervals2d():
    with pytest.raises(ValueError, match="Dimensions of intervals must be 2D."):
        xugrid.Ugrid2d.from_structured_intervals2d(
            x_intervals=[0.0, 1.0, 2.0],
            y_intervals=[2.0, 1.0, 0.0],
        )
    with pytest.raises(ValueError, match="Interval shapes must match."):
        xugrid.Ugrid2d.from_structured_intervals2d(
            x_intervals=[[0.0, 1.0, 2.0]],
            y_intervals=[[2.0, 1.0, 0.0, 4.0]],
        )

    grid = xugrid.Ugrid2d.from_structured_intervals2d(
        x_intervals=[[0.0, 1.0, 2.0], [0.0, 1.0, 2.0], [0.0, 1.0, 2.0]],
        y_intervals=[
            [2.0, 1.0, 0.0],
            [2.0, 1.0, 0.0],
            [2.0, 1.0, 0.0],
        ],
    )
    assert isinstance(grid, xugrid.Ugrid2d)
    assert grid.n_face == 4


def test_from_structured_bounds():
    x_vertices = np.array([1.0, 3.0, 5.0, 7.0, 9.0])
    y_vertices = np.array([2.5, 7.5, 12.5, 17.5])
    # Ascending
    x_bounds = np.column_stack((x_vertices[:-1], x_vertices[1:]))
    y_bounds = np.column_stack((y_vertices[:-1], y_vertices[1:]))
    grid = xugrid.Ugrid2d.from_structured_bounds(x_bounds, y_bounds)
    assert isinstance(grid, xugrid.Ugrid2d)
    assert grid.n_face == 12

    x_bounds = np.array(
        [[[0.0, 0.0, 1.0, 1.0], [2.0, 2.0, 3.0, 3.0], [4.0, 4.0, 5.0, 5.0]]]
    )
    y_bounds = np.array(
        [[[0.0, 1.0, 1.0, 0.0], [2.0, 3.0, 3.0, 2.0], [4.0, 5.0, 5.0, 4.0]]]
    )
    grid = xugrid.Ugrid2d.from_structured_bounds(x_bounds, y_bounds)
    assert grid.n_face == 3
    assert np.allclose(grid.area, 1.0)
    assert grid.bounds == (0.0, 0.0, 5.0, 5.0)

    with pytest.raises(ValueError, match="Bounds shapes do not match"):
        xugrid.Ugrid2d.from_structured_bounds(x_bounds, y_bounds.T)


def test_from_structured():
    da = xr.DataArray(
        data=np.ones((2, 2)),
        coords={"y": [12.0, 11.0], "x": [1.0, 2.0]},
        dims=("y", "x"),
    )
    grid = xugrid.Ugrid2d.from_structured(da)
    assert isinstance(grid, xugrid.Ugrid2d)
    assert grid.n_face == 4

    da = xr.DataArray(
        data=np.ones((2, 2)),
        coords={"lat": [12.0, 11.0], "lon": [1.0, 2.0]},
        dims=("lat", "lon"),
    )
    grid = xugrid.Ugrid2d.from_structured(da, x="lon", y="lat")
    assert isinstance(grid, xugrid.Ugrid2d)
    assert grid.n_face == 4


def test_from_structured_multicoord():
    da = xr.DataArray(
        data=np.ones((2, 2)),
        coords={
            "yc": (("y", "x"), [[12.0, 11.0], [13.0, 12.0]]),
            "xc": (("y", "x"), [[1.0, 2.0], [2.0, 3.0]]),
        },
        dims=("y", "x"),
    )
    grid = xugrid.Ugrid2d._from_structured_multicoord(da, x="xc", y="yc")
    assert isinstance(grid, xugrid.Ugrid2d)
    assert grid.n_face == 4


def test_from_shapely():
    with pytest.raises(TypeError):
        x = np.array([0.0, 1.0, 2.0])
        y = np.array([0.0, 0.0, 0.0])
        xugrid.Ugrid2d.from_shapely(geometry=shapely.linestrings(x, y))

    xy = np.array(
        [
            [0.0, 0.0],
            [1.0, 0.0],
            [1.0, 1.0],
            [0.0, 0.0],
        ]
    )
    grid = xugrid.Ugrid2d.from_shapely(geometry=[shapely.polygons(xy)])
    assert isinstance(grid, xugrid.Ugrid2d)


def test_from_geodataframe():
    with pytest.raises(TypeError, match="Expected GeoDataFrame"):
        xugrid.Ugrid2d.from_geodataframe(1)

    xy = np.array(
        [
            [0.0, 0.0],
            [1.0, 0.0],
            [1.0, 1.0],
            [0.0, 0.0],
        ]
    )
    gdf = gpd.GeoDataFrame(geometry=[shapely.polygons(xy)])
    grid = xugrid.Ugrid2d.from_geodataframe(gdf)
    assert isinstance(grid, xugrid.Ugrid2d)


def test_bounding_polygon():
    grid = grid2d()
    polygon = grid.bounding_polygon()
    assert isinstance(polygon, shapely.Polygon)
    assert np.allclose(grid.bounds, polygon.bounds)


def test_to_shapely():
    grid = grid2d()

    points = grid.to_shapely(f"{NAME}_nNodes")
    assert isinstance(points[0], shapely.Geometry)

    lines = grid.to_shapely(f"{NAME}_nEdges")
    assert isinstance(lines[0], shapely.Geometry)

    polygons = grid.to_shapely(f"{NAME}_nFaces")
    assert isinstance(polygons[0], shapely.Geometry)


def test_grid_from_geodataframe():
    with pytest.raises(TypeError, match="Cannot convert a list"):
        xugrid.conversion.grid_from_geodataframe([])

    with pytest.raises(ValueError, match="geodataframe contains no geometry"):
        xugrid.conversion.grid_from_geodataframe(gpd.GeoDataFrame(geometry=[]))

    x = np.array([0.0, 1.0, 2.0])
    y = np.array([0.0, 0.0, 0.0])
    line = shapely.linestrings(x, y)
    xy = np.array(
        [
            [0.0, 0.0],
            [1.0, 0.0],
            [1.0, 1.0],
            [0.0, 0.0],
        ]
    )
    polygon = shapely.polygons(xy)
    points = shapely.points(x, y)

    with pytest.raises(ValueError, match="Multiple geometry types detected"):
        xugrid.conversion.grid_from_geodataframe(
            gpd.GeoDataFrame(geometry=[line, polygon])
        )
    with pytest.raises(ValueError, match="Invalid geometry type"):
        xugrid.conversion.grid_from_geodataframe(gpd.GeoDataFrame(geometry=points))

    grid = xugrid.conversion.grid_from_geodataframe(gpd.GeoDataFrame(geometry=[line]))
    assert isinstance(grid, xugrid.Ugrid1d)
    grid = xugrid.conversion.grid_from_geodataframe(
        gpd.GeoDataFrame(geometry=[polygon])
    )
    assert isinstance(grid, xugrid.Ugrid2d)


def test_ugrid2d_plot():
    grid = grid2d()
    primitive = grid.plot()
    assert isinstance(primitive, LineCollection)


def test_ugrid2d_rename():
    grid = grid2d()
    original_indexes = grid._indexes.copy()
    original_attrs = grid._attrs.copy()

    renamed = grid.rename("__renamed")

    # Check that original is unchanged
    assert grid._attrs == original_attrs
    assert grid._indexes == original_indexes
    assert renamed._attrs == {
        "cf_role": "mesh_topology",
        "long_name": "Topology data of 2D mesh",
        "topology_dimension": 2,
        "node_dimension": "__renamed_nNodes",
        "edge_dimension": "__renamed_nEdges",
        "face_dimension": "__renamed_nFaces",
        "max_face_nodes_dimension": "__renamed_nMax_face_nodes",
        "boundary_edge_dimension": "__renamed_nBoundary_edges",
        "edge_node_connectivity": "__renamed_edge_nodes",
        "face_node_connectivity": "__renamed_face_nodes",
        "face_edge_connectivity": "__renamed_face_edges",
        "edge_face_connectivity": "__renamed_edge_faces",
        "boundary_node_connectivity": "__renamed_boundary_nodes",
        "face_face_connectivity": "__renamed_face_faces",
        "node_coordinates": "__renamed_node_x __renamed_node_y",
        "edge_coordinates": "__renamed_edge_x __renamed_edge_y",
        "face_coordinates": "__renamed_face_x __renamed_face_y",
    }
    assert renamed._indexes == {
        "node_x": "__renamed_node_x",
        "node_y": "__renamed_node_y",
    }
    assert renamed.name == "__renamed"


def test_ugrid2d_rename_with_dataset():
    grid = grid2d()
    grid2 = xugrid.Ugrid2d.from_dataset(grid.to_dataset())
    original_dataset = grid2._dataset.copy()

    renamed2 = grid2.rename("__renamed")
    dataset = renamed2._dataset
    assert grid2._dataset.equals(original_dataset)
    assert sorted(dataset.data_vars) == [
        "__renamed",
        "__renamed_edge_nodes",
        "__renamed_face_nodes",
    ]
    assert sorted(dataset.dims) == [
        "__renamed_nEdges",
        "__renamed_nFaces",
        "__renamed_nMax_face_nodes",
        "__renamed_nNodes",
        "two",
    ]
    assert sorted(dataset.coords) == ["__renamed_node_x", "__renamed_node_y"]


class TestPeriodicGridConversion:
    @pytest.fixture(autouse=True)
    def setup(self):
        self.vertices = np.array(
            [
                [0.0, 0.0],  # 0
                [1.0, 0.0],  # 1
                [2.0, 0.0],  # 2
                [3.0, 0.0],  # 3
                [0.0, 1.0],  # 4
                [1.0, 1.0],  # 5
                [2.0, 1.0],  # 6
                [3.0, 1.0],  # 7
                [0.0, 2.0],  # 8
                [1.0, 2.0],  # 9
                [2.0, 2.0],  # 10
                [3.0, 2.0],  # 11
            ]
        )
        self.faces = np.array(
            [
                [0, 1, 5, 4],
                [1, 2, 6, 5],
                [2, 3, 7, 6],
                [4, 5, 9, 8],
                [5, 6, 10, 9],
                [6, 7, 11, 10],
            ]
        )
        grid = xugrid.Ugrid2d(*self.vertices.T, -1, self.faces)
        ds = xr.Dataset()
        ds["a"] = xr.DataArray(np.arange(grid.n_node), dims=(grid.node_dimension,))
        ds["b"] = xr.DataArray(np.arange(grid.n_edge), dims=(grid.edge_dimension,))
        ds["c"] = xr.DataArray(np.arange(grid.n_face), dims=(grid.face_dimension,))
        self.ds = ds
        self.grid = grid

    def test_to_periodic(self):
        grid = self.grid.copy()

        # Trigger edge node connectivity
        _ = grid.edge_node_connectivity
        # Convert
        new, new_ds = grid.to_periodic(obj=self.ds)

        # Absent vertices: 3, 7, 11
        expected_vertices = self.vertices[[0, 1, 2, 4, 5, 6, 8, 9, 10]]
        expected_faces = np.array(
            [
                [0, 1, 4, 3],
                [1, 2, 5, 4],
                [2, 0, 3, 5],
                [3, 4, 7, 6],
                [4, 5, 8, 7],
                [5, 3, 6, 8],
            ]
        )
        expected_edges = np.array(
            [
                [0, 1],
                [0, 3],
                [1, 2],
                [1, 4],
                [0, 2],
                [2, 5],
                [3, 4],
                [3, 6],
                [4, 5],
                [4, 7],
                [3, 5],
                [5, 8],
                [6, 7],
                [7, 8],
                [6, 8],
            ]
        )
        assert np.array_equal(
            new.face_node_connectivity,
            expected_faces,
        )
        assert np.allclose(
            new.node_coordinates,
            expected_vertices,
        )
        assert np.array_equal(
            new.edge_node_connectivity,
            expected_edges,
        )
        # Remove nodes (3 & 7 & 11) and edges (6 & 13)
        expected_a = np.arange(grid.n_node).tolist()
        expected_a.remove(3)
        expected_a.remove(7)
        expected_a.remove(11)
        expected_b = np.arange(grid.n_edge).tolist()
        expected_b.remove(6)
        expected_b.remove(13)
        assert np.array_equal(new_ds["a"], expected_a)
        assert np.array_equal(new_ds["b"], expected_b)
        assert np.array_equal(new_ds["c"], [0, 1, 2, 3, 4, 5])

        # Test whether it also works without an object provided.
        new = grid.to_periodic()
        assert np.array_equal(
            new.face_node_connectivity,
            expected_faces,
        )
        assert np.allclose(new.node_coordinates, expected_vertices)
        assert np.array_equal(new.edge_node_connectivity, expected_edges)

    def test_to_nonperiodic(self):
        grid = self.grid.copy()
        _ = grid.edge_node_connectivity  # trigger generation of edge nodes
        periodic_grid, new_ds = grid.to_periodic(obj=self.ds)
        back = periodic_grid.to_nonperiodic(xmax=3.0)

        expected_vertices = self.vertices[[0, 1, 2, 4, 5, 6, 8, 9, 10, 3, 7, 11]]
        expected_faces = np.array(
            [
                [0, 1, 4, 3],
                [1, 2, 5, 4],
                [2, 9, 10, 5],
                [3, 4, 7, 6],
                [4, 5, 8, 7],
                [5, 10, 11, 8],
            ]
        )
        back, back_ds = periodic_grid.to_nonperiodic(xmax=3.0, obj=new_ds)
        assert np.allclose(back.node_coordinates, expected_vertices)
        assert np.array_equal(back.face_node_connectivity, expected_faces)
        assert back.edge_node_connectivity.shape == (17, 2)
        assert np.array_equal(back_ds["a"], [0, 1, 2, 4, 5, 6, 8, 9, 10, 0, 4, 8])
        assert np.array_equal(
            back_ds["b"], [0, 1, 2, 3, 3, 4, 5, 7, 8, 9, 10, 10, 11, 12, 14, 15, 16]
        )
        assert np.array_equal(back_ds["c"], [0, 1, 2, 3, 4, 5])

        back = periodic_grid.to_nonperiodic(xmax=3.0)
        assert np.allclose(back.node_coordinates, expected_vertices)
        assert np.array_equal(back.face_node_connectivity, expected_faces)
        assert back.edge_node_connectivity.shape == (17, 2)


def test_equals():
    grid = grid2d()
    grid_copy = grid2d()
    assert grid.equals(grid)
    assert grid.equals(grid_copy)
    xr_grid = grid.to_dataset()
    assert not grid.equals(xr_grid)
    grid_copy._attrs["attr"] = "something_else"
    assert not grid.equals(grid_copy)


def test_earcut_triangulate_polygons():
    with pytest.raises(TypeError):
        xugrid.Ugrid2d.earcut_triangulate_polygons("abc")

    x = np.array([0.0, 1.0, 2.0])
    y = np.array([0.0, 0.0, 0.0])
    gdf = gpd.GeoDataFrame(geometry=[shapely.linestrings(x, y)])
    with pytest.raises(TypeError):
        xugrid.Ugrid2d.earcut_triangulate_polygons(gdf)

    xy = np.array(
        [
            [0.0, 0.0],
            [1.0, 0.0],
            [1.0, 1.0],
            [0.0, 1.0],
            [0.0, 0.0],
        ]
    )
    hole = np.array(
        [
            [
                [0.25, 0.25],
                [0.75, 0.25],
                [0.75, 0.75],
                [0.25, 0.25],
            ]
        ]
    )
    polygon = shapely.polygons(xy, holes=hole)
    gdf = gpd.GeoDataFrame(geometry=[polygon])

    grid = xugrid.Ugrid2d.earcut_triangulate_polygons(polygons=gdf)
    assert isinstance(grid, xugrid.Ugrid2d)
    assert np.issubdtype(grid.face_node_connectivity.dtype, np.signedinteger)
    assert np.allclose(polygon.area, grid.area.sum())

    grid, index = xugrid.Ugrid2d.earcut_triangulate_polygons(
        polygons=gdf, return_index=True
    )
    assert isinstance(grid, xugrid.Ugrid2d)
    assert isinstance(index, np.ndarray)
    assert (index == 0).all()


def test_ugrid2d_create_data_array():
    grid = grid2d()

    uda = grid.create_data_array(np.zeros(grid.n_node), facet="node")
    assert isinstance(uda, xugrid.UgridDataArray)

    uda = grid.create_data_array(np.zeros(grid.n_edge), facet="edge")
    assert isinstance(uda, xugrid.UgridDataArray)

    uda = grid.create_data_array(np.zeros(grid.n_face), facet="face")
    assert isinstance(uda, xugrid.UgridDataArray)

    # Error on facet
    with pytest.raises(ValueError, match="Invalid facet"):
        grid.create_data_array([1, 2, 3, 4], facet="volume")

    # Error on on dimensions
    with pytest.raises(ValueError, match="Can only create DataArrays from 1D arrays"):
        grid.create_data_array([[1, 2, 3, 4]], facet="face")

    # Error on size
    with pytest.raises(ValueError, match="Conflicting sizes"):
        grid.create_data_array([1, 2, 3, 4, 5], facet="face")


def test_ugrid2d_format_connectivity():
    grid = grid2d()
    assert isinstance(grid.face_node_connectivity, np.ndarray)
    assert isinstance(
        grid.format_connectivity_as_sparse(grid.face_node_connectivity),
        sparse.csr_matrix,
    )
    assert isinstance(grid.node_node_connectivity, sparse.csr_matrix)
    assert isinstance(
        grid.format_connectivity_as_dense(grid.node_node_connectivity), np.ndarray
    )
    assert isinstance(
        grid.format_connectivity_as_sparse(grid.node_node_connectivity.tocoo()),
        sparse.csr_matrix,
    )


def test_nearest_interpolate():
    node_x = np.array([-0.5, -0.5, 0.5, 0.5, 1.5, 1.5, 2.5, 2.5, 3.5, 3.5, 4.5, 4.5])
    node_y = np.array(
        [-1.0, 1.0, 1.0, -1.0, 1.0, -1.0, 1.0, -1.0, 1.0, -1.0, 1.0, -1.0]
    )
    face_node_connectivity = np.array(
        [
            [0, 1, 2, 3],  # First quad
            [3, 2, 4, 5],  # Second quad
            [5, 4, 6, 7],  # Third quad
            [7, 6, 8, 9],  # Fourth quad
            [9, 8, 10, 11],  # Fifth quad
        ]
    )
    grid = xugrid.Ugrid2d(node_x, node_y, -1, face_node_connectivity)

    # Centroids are:
    # x = np.array([0.0, 1.0, 2.0, 3.0, 4.0])
    # y = np.zeros_like(x)
    data = np.array([0.0, np.nan, np.nan, np.nan, 4.0])
    facedim = grid.face_dimension
    actual = grid._nearest_interpolate(data, facedim, np.inf)
    assert np.allclose(actual, np.array([0.0, 0.0, 0.0, 4.0, 4.0]))

    actual = grid._nearest_interpolate(data, facedim, 1.1)
    assert np.allclose(actual, np.array([0.0, 0.0, np.nan, 4.0, 4.0]), equal_nan=True)

    with pytest.raises(ValueError, match="All values are NA."):
        grid._nearest_interpolate(np.full_like(data, np.nan), facedim, np.inf)

    # Test for node_dimension
    data = np.full(grid.n_node, np.nan)
    data[0] = 1.0
    actual = grid._nearest_interpolate(data, grid.node_dimension, np.inf)
    assert np.allclose(actual, 1.0)

    # Test for edge_dimension
    data = np.full(grid.n_edge, np.nan)
    data[0] = 1.0
    actual = grid._nearest_interpolate(data, grid.edge_dimension, np.inf)
    assert np.allclose(actual, 1.0)


def test_locate_nearest():
    grid = grid2d()

    indices = grid.locate_nearest_node(grid.node_coordinates)
    assert np.array_equal(indices, [0, 1, 2, 3, 4, 5, 6])

    indices = grid.locate_nearest_edge(grid.edge_coordinates)
    assert np.array_equal(indices, [0, 1, 2, 3, 4, 5, 6, 7, 8, 9])

    indices = grid.locate_nearest_face(grid.face_coordinates)
    assert np.array_equal(indices, [0, 1, 2, 3])

    assert np.array_equal(grid.locate_nearest_node([[-10.0, 0.0]], 1.0), [-1])
    assert np.array_equal(grid.locate_nearest_edge([[-10.0, 0.0]], 1.0), [-1])
    assert np.array_equal(grid.locate_nearest_face([[-10.0, 0.0]], 1.0), [-1])



import matplotlib.pyplot as plt
import numpy as np
import pytest
import xarray as xr
from matplotlib.collections import LineCollection, PathCollection, PolyCollection
from matplotlib.image import AxesImage
from matplotlib.tri import TriContourSet

import xugrid
from xugrid.plot import plot


class TestPlot:
    @pytest.fixture(autouse=True)
    def setup(self):
        # Remove all matplotlib figures
        plt.close("all")

        self.ds = xugrid.data.disk()
        self.grid = self.ds.ugrid.grids[0]
        self.node = self.ds["node_z"]
        self.edge = self.ds["edge_z"]
        self.face = self.ds["face_z"]
        self.node_da = self.node.ugrid.obj
        self.edge_da = self.edge.ugrid.obj
        self.face_da = self.face.ugrid.obj

    def test_get_ugrid_dim(self):
        with pytest.raises(ValueError, match="Not a valid UGRID dimension"):
            da = xr.DataArray([0, 1], dims=["x"])
            plot.get_ugrid_dim(self.grid, da)

        assert plot.get_ugrid_dim(self.grid, self.node_da) == plot.NODE
        assert plot.get_ugrid_dim(self.grid, self.edge_da) == plot.EDGE
        assert plot.get_ugrid_dim(self.grid, self.face_da) == plot.FACE

    def test_plot_contour(self):
        with pytest.raises(ValueError, match="contour only supports"):
            plot.contour(self.grid, self.edge_da)
        with pytest.raises(ValueError, match="contour only supports"):
            self.edge.ugrid.plot.contour()

        assert isinstance(plot.contour(self.grid, self.node_da), TriContourSet)
        assert isinstance(self.node.ugrid.plot.contour(), TriContourSet)
        assert isinstance(plot.contour(self.grid, self.face_da), TriContourSet)
        assert isinstance(self.face.ugrid.plot.contour(), TriContourSet)

        # Check whether line is dashed or not?
        assert isinstance(self.face.ugrid.plot.contour(colors="black"), TriContourSet)

    def test_plot_contourf(self):
        with pytest.raises(ValueError, match="contourf only supports"):
            plot.contourf(self.grid, self.edge_da)
        with pytest.raises(ValueError, match="contourf only supports"):
            self.edge.ugrid.plot.contourf()

        assert isinstance(plot.contourf(self.grid, self.node_da), TriContourSet)
        assert isinstance(self.node.ugrid.plot.contourf(), TriContourSet)
        assert isinstance(plot.contourf(self.grid, self.face_da), TriContourSet)
        assert isinstance(self.face.ugrid.plot.contourf(), TriContourSet)

    def test_plot_imshow(self):
        with pytest.raises(ValueError, match="imshow only supports"):
            plot.imshow(self.grid, self.edge_da)
        with pytest.raises(ValueError, match="imshow only supports"):
            self.edge.ugrid.plot.imshow()

        with pytest.raises(ValueError, match="imshow only supports"):
            plot.imshow(self.grid, self.node_da)
        with pytest.raises(ValueError, match="imshow only supports"):
            self.node.ugrid.plot.imshow()

        # Reduce resolution for non-JITed test runs (e.g. for coverage)
        assert isinstance(
            plot.imshow(self.grid, self.face_da, resolution=1.0), AxesImage
        )
        assert isinstance(self.face.ugrid.plot.imshow(resolution=1.0), AxesImage)

        with pytest.raises(ValueError, match="kwarg is not available in xugrid"):
            self.face.ugrid.plot.imshow(resolution=1.0, size=4, aspect="equal")

        actual = plot.imshow(
            self.grid, self.face_da, resolution=1.0, extent=(0.0, 1.0, 0.0, 1.0)
        )
        assert isinstance(actual, AxesImage)
        actual = plot.imshow(
            self.grid,
            self.face_da,
            resolution=1.0,
            extent=(0.0, 1.0, 0.0, 1.0),
            origin="upper",
        )
        assert isinstance(actual, AxesImage)
        actual = plot.imshow(
            self.grid,
            self.face_da,
            resolution=1.0,
            extent=(0.0, 1.0, 1.0, 0.0),
            origin="lower",
        )
        assert isinstance(actual, AxesImage)

    def test_plot_line(self):
        with pytest.raises(ValueError, match="line only supports"):
            plot.line(self.grid, self.node_da)
        with pytest.raises(ValueError, match="line only supports"):
            plot.line(self.grid, self.face_da)

        assert isinstance(plot.line(self.grid), LineCollection)
        assert isinstance(plot.line(self.grid, self.edge_da), LineCollection)
        assert isinstance(self.node.ugrid.plot.line(), LineCollection)
        assert isinstance(self.edge.ugrid.plot.line(), LineCollection)
        assert isinstance(self.face.ugrid.plot.line(), LineCollection)

    def test_plot_pcolormesh(self):
        with pytest.raises(ValueError, match="pcolormesh only supports"):
            plot.pcolormesh(self.grid, self.edge_da)
        with pytest.raises(ValueError, match="pcolormesh only supports"):
            self.edge.ugrid.plot.pcolormesh()

        with pytest.raises(ValueError, match="pcolormesh only supports"):
            plot.pcolormesh(self.grid, self.node_da)
        with pytest.raises(ValueError, match="pcolormesh only supports"):
            self.node.ugrid.plot.pcolormesh()

        assert isinstance(plot.pcolormesh(self.grid, self.face_da), PolyCollection)
        assert isinstance(self.face.ugrid.plot.pcolormesh(), PolyCollection)

    def test_plot_surface(self):
        with pytest.raises(ValueError, match="surface only supports"):
            plot.surface(self.grid, self.edge_da)
        with pytest.raises(ValueError, match="surface only supports"):
            self.edge.ugrid.plot.surface()

        assert isinstance(plot.surface(self.grid, self.node_da), PolyCollection)
        assert isinstance(plot.surface(self.grid, self.face_da), PolyCollection)
        assert isinstance(self.node.ugrid.plot.surface(), PolyCollection)
        assert isinstance(self.face.ugrid.plot.surface(), PolyCollection)

        with pytest.raises(ValueError, match="If ax is passed to surface()"):
            _, ax = plt.subplots()
            self.face.ugrid.plot.surface(ax=ax)

        fig = plt.figure()
        ax = fig.add_subplot(projection="3d")
        assert isinstance(self.face.ugrid.plot.surface(ax=ax), PolyCollection)

    def test_plot_scatter(self):
        assert isinstance(plot.scatter(self.grid, self.node_da), PathCollection)
        assert isinstance(plot.scatter(self.grid, self.edge_da), PathCollection)
        assert isinstance(plot.scatter(self.grid, self.face_da), PathCollection)
        assert isinstance(self.node.ugrid.plot.scatter(), PathCollection)
        assert isinstance(self.edge.ugrid.plot.scatter(), PathCollection)
        assert isinstance(self.face.ugrid.plot.scatter(), PathCollection)

    def test_plot_tripcolor(self):
        with pytest.raises(ValueError, match="tripcolor only supports"):
            plot.tripcolor(self.grid, self.edge_da)
        with pytest.raises(ValueError, match="tripcolor only supports"):
            self.edge.ugrid.plot.tripcolor()

        with pytest.raises(ValueError, match="tripcolor only supports"):
            plot.tripcolor(self.grid, self.face_da)
        with pytest.raises(ValueError, match="tripcolor only supports"):
            self.face.ugrid.plot.tripcolor()

        assert isinstance(plot.tripcolor(self.grid, self.node_da), PolyCollection)
        assert isinstance(self.node.ugrid.plot.tripcolor(), PolyCollection)

    def test_plot(self):
        assert isinstance(self.node.ugrid.plot(), PolyCollection)
        assert isinstance(self.edge.ugrid.plot(), LineCollection)
        assert isinstance(self.face.ugrid.plot(), PolyCollection)

    def test_add_colorbar_exception(self):
        with pytest.raises(ValueError, match="cbar_ax and cbar_kwargs"):
            cbar_kwargs = {"location": "top"}
            self.face.ugrid.plot(add_colorbar=False, cbar_kwargs=cbar_kwargs)

    def test_default_size(self):
        with pytest.raises(ValueError, match="cannot provide both `size` and `ax`"):
            _, ax = plt.subplots()
            self.face.ugrid.plot(ax=ax, size=10)

        self.face.ugrid.plot(size=10, add_colorbar=False)
        fig = plt.gca().figure
        h = fig.get_figheight()
        w = fig.get_figwidth()
        assert np.allclose(w / h, 1.0)

        self.face.ugrid.plot(size=10)
        fig = plt.gca().figure
        h = fig.get_figheight()
        w = fig.get_figwidth()
        assert np.allclose(w / h, 1.26)

    def test_error_dimension(self):
        with pytest.raises(
            ValueError, match="UgridDataArray contains non-topology dimensions"
        ):
            uda = xugrid.concat(
                [
                    self.face.assign_coords(time="2000-01-01"),
                    self.face.assign_coords(time="2001-01-01"),
                ],
                dim="time",
            )
            uda.ugrid.plot()



from __future__ import annotations

import warnings
from typing import Any, Callable, Dict, NamedTuple, Tuple

import numba as nb
import numpy as np
import pandas as pd
import xarray as xr
from scipy import sparse

from xugrid.constants import FloatArray, IntArray
from xugrid.core.sparse import MatrixCSR, columns_and_values, nzrange, row_slice


@nb.njit(inline="always")
def lower_slice(ilu, row: int) -> slice:
    return slice(ilu.indptr[row], ilu.uptr[row])


@nb.njit(inline="always")
def upper_slice(ilu, row: int) -> slice:
    return slice(ilu.uptr[row], ilu.indptr[row + 1])


@nb.njit
def set_uptr(ilu: ILU0Preconditioner) -> None:
    # i is row index, j is column index
    for i in range(ilu.n):
        for nzi in nzrange(ilu, i):
            j = ilu.indices[nzi]
            if j > i:
                ilu.uptr[i] = nzi
                break
    return


@nb.njit
def _update(ilu: ILU0Preconditioner, A: MatrixCSR, delta: float, relax: float):
    """
    Perform zero fill-in incomplete lower-upper (ILU0) factorization
    using the values of A.
    """
    ilu.work[:] = 0.0
    visited = np.full(ilu.n, False)

    # i is row index, j is column index, v is value.
    for i in range(ilu.n):
        for j, v in columns_and_values(A, row_slice(A, i)):
            visited[j] = True
            ilu.work[j] += v

        rs = 0.0
        for j in ilu.indices[lower_slice(ilu, i)]:
            # Compute row multiplier
            multiplier = ilu.work[j] * ilu.diagonal[j]
            ilu.work[j] = multiplier
            # Perform linear combination
            for jj, vv in columns_and_values(ilu, upper_slice(ilu, j)):
                if visited[jj]:
                    ilu.work[jj] -= multiplier * vv
                else:
                    rs += multiplier * vv

        diag = ilu.work[i]
        multiplier = (1.0 + delta) * diag - (relax * rs)
        # Work around a zero-valued pivot and make sure the multiplier hasn't
        # changed sign.
        if multiplier == 0:
            multiplier = 1e-6
        elif np.sign(multiplier) != np.sign(diag):
            multiplier = np.sign(diag) * 1.0e-6
        ilu.diagonal[i] = 1.0 / multiplier

        # Reset work arrays, assign off-diagonal values
        visited[i] = False
        ilu.work[i] = 0.0
        for nzi in nzrange(ilu, i):
            j = ilu.indices[nzi]
            ilu.data[nzi] = ilu.work[j]
            ilu.work[j] = 0.0
            visited[j] = False

    return


@nb.njit
def _solve(ilu: ILU0Preconditioner, r: np.ndarray):
    r"""
    LU \ r

    Stores the result in the pre-allocated work array.
    """
    ilu.work[:] = 0.0

    # forward
    for i in range(ilu.n):
        value = r[i]
        for j, v in columns_and_values(ilu, lower_slice(ilu, i)):
            value -= v * ilu.work[j]
        ilu.work[i] = value

    # backward
    for i in range(ilu.n - 1, -1, -1):
        value = ilu.work[i]
        for j, v in columns_and_values(ilu, upper_slice(ilu, i)):
            value -= v * ilu.work[j]
        ilu.work[i] = value * ilu.diagonal[i]

    return


class ILU0Preconditioner(NamedTuple):
    """
    Preconditioner based on zero fill-in lower-upper (ILU0) factorization.

    Data is stored in compressed sparse row (CSR) format. The diagonal
    values have been extracted for easier access. Upper and lower values
    are stored in CSR format. Next to the indptr array, which identifies
    the start and end of each row, the uptr array has been added to
    identify the start to the right of the diagonal. In case the row to the
    right of the diagonal is empty, it contains the end of the rows as
    indicated by the indptr array.

    Parameters
    ----------
    n: int
        Number of rows
    m: int
        Number of columns
    indptr: np.ndarray of int
        CSR format index pointer array of the matrix
    uptr: np.ndarray of int
        CSR format index pointer array of the upper elements (diagonal or higher)
    indices: np.ndarray of int
        CSR format index array of the matrix
    data: np.ndarray of float
        CSR format data array of the matrix
    diagonal: np.ndarray of float
        Diagonal values of LU factorization
    work: np.ndarray of float
        Work array. Used in factorization and solve.
    """

    n: int
    m: int
    indptr: IntArray
    uptr: IntArray
    indices: IntArray
    data: FloatArray
    diagonal: FloatArray
    work: FloatArray

    @property
    def shape(self) -> Tuple[int, int]:
        return (self.n, self.m)

    @property
    def dtype(self):
        return self.data.dtype

    @staticmethod
    def from_csr_matrix(
        A: sparse.csr_matrix, delta: float = 0.0, relax: float = 0.0
    ) -> "ILU0Preconditioner":
        # Create a copy of the sparse matrix with the diagonals removed.
        n, m = A.shape
        coo = A.tocoo()
        i = coo.row
        j = coo.col
        offdiag = i != j
        ii = i[offdiag]
        indices = j[offdiag]
        indptr = sparse.csr_matrix((indices, (ii, indices)), shape=A.shape).indptr

        ilu = ILU0Preconditioner(
            n=n,
            m=m,
            indptr=indptr,
            uptr=indptr[1:].copy(),
            indices=indices,
            data=np.empty(indices.size),
            diagonal=np.empty(n),
            work=np.empty(n),
        )
        set_uptr(ilu)

        _update(ilu, MatrixCSR.from_csr_matrix(A), delta, relax)
        return ilu

    def update(self, A, delta=0.0, relax=0.0) -> None:
        _update(self, MatrixCSR.from_csr_matrix(A), delta, relax)
        return

    def matvec(self, r) -> FloatArray:
        _solve(self, r)
        return self.work

    def __repr__(self) -> str:
        return f"ILU0Preconditioner of type {self.dtype} and shape {self.shape}"


def laplace_interpolate(
    data: FloatArray,
    connectivity: sparse.csr_matrix,
    components_labels: IntArray,
    use_weights: bool,
    direct_solve: bool = False,
    delta=0.0,
    relax=0.0,
    atol: float = 1e-4,
    rtol: float = 0.0,
    maxiter: int = 500,
):
    """
    Fill gaps in ``data`` (``np.nan`` values) using Laplace interpolation.

    This solves Laplace's equation where there is no data, with data values
    functioning as fixed potential boundary conditions.

    Note that an iterative solver method will be required for large grids.
    Refer to the documentation of :py:func:`scipy.sparse.linalg.cg`.

    Parameters
    ----------
    data: ndarray of floats with shape ``(n,)``
    connectivity: scipy.sparse.csr_matrix with shape ``(n, n)``
        Sparse connectivity matrix containing ``n_nonzero`` indices and weight values.
    components_labels: array of int with shape ``(n,)``
        The labels of the connected components, result of
        ``scipy.sparse.csgraph.connected_components``.
    use_weights: bool, default False.
        Wether to use the data attribute of the connectivity matrix as
        coefficients. If ``False``, defaults to uniform coefficients of 1.
    direct_solve: bool, optional, default ``False``
        Whether to use a direct or an iterative solver or a conjugate gradient
        solver. Direct methods provides an exact answer, but are unsuitable
        for large problems.
    delta: float, default 0.0
        ILU0 preconditioner non-diagonally dominant correction.
    relax: float, default 0.0
        Modified ILU0 preconditioner relaxation factor.
    atol: float, optional, default 1.0e-4.
        Convergence tolerance for ``scipy.sparse.linalg.cg``.
    rtol: float, optional, default 0.0.
        Convergence tolerance for ``scipy.sparse.linalg.cg``.
    maxiter: int, default 500.
        Maximum number of iterations for ``scipy.sparse.linalg.cg``.

    Returns
    -------
    filled: ndarray of floats
    """
    # Input checks
    n, m = connectivity.shape
    if n != m:
        raise ValueError(f"connectivity is not a square matrix: ({n}, {m})")
    if data.shape != (n,):
        raise ValueError(f"expected data of shape ({n},), received: {data.shape}")

    # Find the elements with data
    isnull = np.isnan(data)
    notnull = ~isnull
    if isnull.all():
        raise ValueError("data is fully nodata")
    if notnull.all():
        return data.copy()

    # Check for parts without any values; these result in a singular matrix.
    # Each label should have at least one notnull value.
    # In case the entirety is null, keep the NaNs there.
    all_null = (
        pd.DataFrame({"label": components_labels, "isnull": isnull})
        .groupby("label")["isnull"]
        .all()
    ).to_numpy()[components_labels]
    known = notnull & ~all_null
    unknown = isnull & ~all_null

    # Create uniform weighting if sparse data is not to be used.
    W = connectivity.astype(np.float64, copy=True)
    if not use_weights:
        W.data[:] = 1.0

    # Compute the (weighted) degree matrix
    D = np.asarray(W.sum(axis=1)).ravel()
    # Compute the Laplacian
    L = sparse.diags(D) - W

    # Extract knowns and unknowns
    A = L[unknown][:, unknown]
    rhs = -L[unknown][:, known].dot(data[known])

    # Diagonal scaling for better conditioning
    diagA = A.diagonal().copy()
    # Guard against non-positive diagonals just in case:
    diagA[diagA <= 0.0] = 1e-10 * abs(diagA).mean()
    scale = 1.0 / np.sqrt(diagA)
    S = sparse.diags(scale)
    A_scaled = S @ A @ S
    rhs_scaled = scale * rhs

    if direct_solve:
        x = sparse.linalg.spsolve(A_scaled, rhs_scaled)
    else:
        # Create preconditioner M
        M = ILU0Preconditioner.from_csr_matrix(A_scaled, delta=delta, relax=relax)
        # IILU isn't guaranteed to preserve symmetry, which could cause technically
        # cause problems for CG. The ILU0 is a port of MODFLOW 6's preconditioner,
        # which is also combined with CG; it (apparently) often works fine although
        # it's not strictly mathematically pure CG.
        x, info = sparse.linalg.cg(
            A_scaled, rhs_scaled, rtol=rtol, atol=atol, maxiter=maxiter, M=M
        )
        if info < 0:
            raise ValueError("scipy.sparse.linalg.cg: illegal input or breakdown")
        elif info > 0:
            warnings.warn(
                f"Failed to converge after {maxiter} iterations",
                UserWarning,
                stacklevel=2,
            )

    out = data.copy()
    out[unknown] = scale * x
    return out


def interpolate_na_helper(
    da: xr.DataArray,
    ugrid_dim: str,
    func: Callable,
    kwargs: Dict[str, Any],
):
    """Use apply ufunc to broadcast over the non UGRID dims."""
    da_filled = xr.apply_ufunc(
        func,
        da,
        input_core_dims=[[ugrid_dim]],
        output_core_dims=[[ugrid_dim]],
        vectorize=True,
        kwargs=kwargs,
        dask="parallelized",
        keep_attrs=True,
        output_dtypes=[da.dtype],
    )
    return da_filled



import numba
import numpy as np

from xugrid.constants import IntDType
from xugrid.regrid.utils import alt_cumsum


@numba.njit(inline="always")
def minmax(v, lower, upper):
    return min(upper, max(lower, v))


def _find_indices(side):
    @numba.njit(inline="always")
    def lt(x, y):
        return x < y

    @numba.njit(inline="always")
    def le(x, y):
        return x <= y

    if side == "right":
        func = le
        add = -1
    elif side == "left":
        func = lt
        add = 1
    else:
        raise ValueError("side must be left or right")

    @numba.njit
    def searchsorted_inner(a, v, v_last, lo, hi, n):
        """
        Perform inner loop of searchsorted (i.e. a binary search).
        This is loosely based on the NumPy implementation in [1]_.

        Parameters
        ----------
        a: 1-D array_like
            The input array.
        v: array_like
            The current value to insert into `a`.
        v_last: array_like
            The previous value inserted into `a`.
        lo: int
            The initial/previous "low" value of the binary search.
        hi: int
            The initial/previous "high" value of the binary search.
        n: int
            The length of `a`.
        .. [1] https://github.com/numpy/numpy/blob/809e8d26b03f549fd0b812a17b8a166bcd966889/numpy/core/source/npysort/binsearch.cpp#L173
        """  # noqa: E501
        if v_last < v:
            hi = n
        else:
            lo = 0
            hi = hi + 1 if hi < n else n

        while hi > lo:
            mid = (lo + hi) >> 1
            if func(a[mid], v):
                # mid is too low => go up
                lo = mid + 1
            else:
                # mid is too high, or is a NaN => go down
                hi = mid
        return lo

    @numba.njit
    def preallocated_searchsorted(a, v, out, nan_helper, sorter) -> None:
        # source bounds may contain NaN values. Eliminate those here and store
        # the original positions in the sorter.
        jj = 0
        for ii, vv in enumerate(a):
            if np.isnan(vv):
                continue
            nan_helper[jj] = vv
            sorter[jj] = ii
            jj += 1

        n = jj - 1
        lo = 0
        hi = n
        v_last = v[0]
        for i in range(len(v)):
            v_search = v[i]
            if np.isnan(v_search):
                continue
            lo = searchsorted_inner(nan_helper, v_search, v_last, lo, hi, n)
            v_last = v_search
            out[i] = minmax(sorter[lo] + add, 0, a.size)
        return

    @numba.njit
    def find_indices(
        source,
        target,
        source_index,
        target_index,
    ):
        """
        Find the indices of target in source. Allocate the result in slices of
        source_index and target_index.

        This is basically a workaround. Numpy searchsorted does not support an axis
        argument to search one nD array on the other.
        See: https://github.com/numpy/numpy/issues/4224

        Fortunately, numba implements searchsorted here:
        https://github.com/numba/numba/blob/f867999c7453141642ea9af21febef796da9ca93/numba/np/arraymath.py#L3647
        But it continously allocates new result arrays.

        Parameters
        ----------
        source: np.ndarray of shape (n, m)
            All other dims flatted to dimension with size n.
        target: np.ndarray of shape (n, m)
            All other dims flatted to dimension with size n.
        source_index: np.ndarray of shape (n_index,)
        target_index: np.ndarray of shape (n_index,)
        """
        _, n = source.shape
        _, m = target.shape
        # Will contain the index along dimension m.
        indices = np.full((source_index.size, m), -1, dtype=IntDType)
        sorter = np.empty(n, IntDType)
        nan_helper = np.empty(n, source.dtype)
        for k, (i, j) in enumerate(zip(source_index, target_index)):
            source_i = source[i, :]
            target_j = target[j, :]
            preallocated_searchsorted(
                source_i, target_j, indices[k], nan_helper, sorter
            )
        return indices

    return find_indices


find_lower_indices = _find_indices("right")
find_upper_indices = _find_indices("left")


def vectorized_overlap(bounds_a, bounds_b):
    """
    Vectorized overlap computation.

    Compare with:

    overlap = max(0, min(a[1], b[1]) - max(a[0], b[0]))
    """
    return np.maximum(
        0.0,
        np.minimum(bounds_a[:, 1], bounds_b[:, 1])
        - np.maximum(bounds_a[:, 0], bounds_b[:, 0]),
    )


def overlap_1d_nd(
    source_bounds,
    target_bounds,
    source_index,
    target_index,
):
    """
    As this uses a binary search, bounds must be monotonic ascending.

    source_bounds and target_bounds may both contain NaN values to denote NoData.

    Parameters
    ----------
    source_bounds: np.ndarray with shape (n, source_size, 2)
    target_bounds: np.ndarray with shape (m, target_size, 2)
    source_index: np.ndarray with shape (o,)
        Used to index into n axis of source_bounds
    target_index: np.ndarray with shape (o,)
        Used to index into m axis of target_bounds

    Returns
    -------
    flat_source_index: np.ndarray of integers with shape (p,)
    flat_target_index: np.ndarray of integers with shape (p,)
    overlap: np.ndarray of floats with shape (p,)
    """
    # _d refers to the dimension of interest _nd refers to the (variable!)
    # number of other dimensions which are flattened and come before d.

    _, target_d_size, _ = target_bounds.shape

    source_lower = source_bounds[..., 0]
    source_upper = source_bounds[..., 1]
    target_lower = target_bounds[..., 0]
    target_upper = target_bounds[..., 1]

    lower_indices = find_lower_indices(
        source_lower, target_lower, source_index, target_index
    )
    upper_indices = find_upper_indices(
        source_upper, target_upper, source_index, target_index
    )

    n_overlap = upper_indices - lower_indices
    n_overlap_nd = n_overlap.sum(axis=1)
    n_total = n_overlap_nd.sum()

    # Create target index
    target_index_d = np.repeat(
        np.broadcast_to(
            np.arange(target_d_size, dtype=IntDType), target_bounds.shape[:2]
        ).ravel(),
        n_overlap.ravel(),
    )

    # Create source index
    increment = alt_cumsum(np.ones(n_total, dtype=IntDType)) - np.repeat(
        alt_cumsum(n_overlap.ravel()), n_overlap.ravel()
    )
    source_index_d = (
        np.repeat(
            lower_indices.ravel(),
            n_overlap.ravel(),
        )
        + increment
    )

    # Now turn them into a linear index.
    target_linear_index = np.ravel_multi_index(
        (np.repeat(target_index, n_overlap_nd), target_index_d),
        dims=target_bounds.shape[:2],
    )
    source_linear_index = np.ravel_multi_index(
        (np.repeat(source_index, n_overlap_nd), source_index_d),
        dims=source_bounds.shape[:2],
    )

    # Compute overlap.
    overlap = vectorized_overlap(
        source_bounds.reshape((-1, 2))[source_linear_index],
        target_bounds.reshape((-1, 2))[target_linear_index],
    )
    valid = overlap > 0.0
    return (source_linear_index[valid], target_linear_index[valid], overlap[valid])


def overlap_1d(
    source_bounds,
    target_bounds,
):
    return overlap_1d_nd(
        source_bounds[np.newaxis],
        target_bounds[np.newaxis],
        np.array([0]),
        np.array([0]),
    )



import numpy as np

from xugrid.regrid import overlap_1d


def test_minmax():
    assert overlap_1d.minmax(-1.0, 0.0, 2.0) == 0.0
    assert overlap_1d.minmax(3.0, 0.0, 2.0) == 2.0
    assert overlap_1d.minmax(1.0, 0.0, 2.0) == 1.0


def test_find_indices():
    a = np.arange(0.0, 11.0)[np.newaxis, :]
    b = np.arange(0.0, 12.5, 2.5)[np.newaxis, :]
    source_index = np.array([0])
    target_index = np.array([0])

    lower = overlap_1d.find_lower_indices(a, b, source_index, target_index)
    upper = overlap_1d.find_upper_indices(a, b, source_index, target_index)
    assert np.array_equal(lower, [[0, 2, 5, 7, 9]])
    assert np.array_equal(upper, [[1, 4, 6, 9, 11]])


def test_vectorized_overlap():
    bounds_a = np.array(
        [
            [0.0, 3.0],
            [0.0, 3.0],
        ]
    )
    bounds_b = np.array(
        [
            [1.0, 2.0],
            [1.0, 2.0],
        ]
    )
    actual = overlap_1d.vectorized_overlap(bounds_a, bounds_b)
    assert np.array_equal(actual, np.array([1.0, 1.0]))


def test_overlap_1d():
    source_bounds = np.array(
        [
            [0.0, 1.0],
            [2.0, 3.0],
            [np.nan, np.nan],
            [5.0, 6.0],
        ]
    )
    target_bounds = np.array(
        [
            [0.0, 10.0],
            [10.0, 20.0],
        ]
    )
    source, target, overlap = overlap_1d.overlap_1d(source_bounds, target_bounds)
    assert np.array_equal(source, [0, 1, 3])
    assert np.array_equal(target, [0, 0, 0])
    assert np.allclose(overlap, [1.0, 1.0, 1.0])

    target_bounds = np.array(
        [
            [0.0, 2.5],
            [np.nan, np.nan],
        ]
    )
    source, target, overlap = overlap_1d.overlap_1d(source_bounds, target_bounds)
    assert np.array_equal(source, [0, 1])
    assert np.array_equal(target, [0, 0])
    assert np.allclose(overlap, [1.0, 0.5])


def test_overlap_1d_nd():
    source_bounds = np.array(
        [
            [
                [0.0, 1.0],
                [2.0, 3.0],
                [np.nan, np.nan],
                [5.0, 6.0],
            ]
        ]
    )
    target_bounds = np.array(
        [
            [
                [0.0, 10.0],
                [10.0, 20.0],
            ],
            [
                [0.0, 2.5],
                [np.nan, np.nan],
            ],
        ]
    )
    source_index = np.array([0, 0])
    target_index = np.array([0, 1])
    source, target, overlap = overlap_1d.overlap_1d_nd(
        source_bounds, target_bounds, source_index, target_index
    )
    assert np.array_equal(source, [0, 1, 3, 0, 1])
    assert np.array_equal(target, [0, 0, 0, 2, 2])
    assert np.allclose(overlap, [1.0, 1.0, 1.0, 1.0, 0.5])



import geopandas as gpd
import numpy as np
import pytest
import shapely
from shapely.geometry import Polygon

import xugrid as xu
from xugrid.ugrid import burn


@pytest.fixture(scope="function")
def grid():
    """Three by three squares"""
    x = np.arange(0.0, 4.0)
    y = np.arange(0.0, 4.0)
    node_y, node_x = [a.ravel() for a in np.meshgrid(y, x, indexing="ij")]
    nx = ny = 3
    # Define the first vertex of every face, v.
    v = (np.add.outer(np.arange(nx), nx * np.arange(ny)) + np.arange(ny)).T.ravel()
    faces = np.column_stack((v, v + 1, v + nx + 2, v + nx + 1))
    return xu.Ugrid2d(node_x, node_y, -1, faces)


@pytest.fixture(scope="function")
def points_and_values():
    xy = np.array(
        [
            [0.5, 0.5],
            [1.5, 0.5],
            [2.5, 2.5],
        ]
    )
    points = gpd.points_from_xy(*xy.T)
    values = np.array([0.0, 1.0, 3.0])
    return points, values


@pytest.fixture(scope="function")
def lines_and_values():
    xy = np.array(
        [
            [0.5, 0.5],
            [2.5, 0.5],
            [1.2, 1.5],
            [1.8, 1.5],
            [0.2, 2.2],
            [0.8, 2.8],
            [1.2, 2.2],
            [1.8, 2.8],
        ]
    )
    indices = np.array([0, 0, 1, 1, 2, 2, 2, 2])
    values = np.array([0, 1, 2])
    lines = gpd.GeoSeries(shapely.linestrings(xy, indices=indices))
    return lines, values


@pytest.fixture(scope="function")
def polygons_and_values():
    values = [0, 1]
    polygons = gpd.GeoSeries(
        [
            shapely.Polygon(shell=[(0.0, 0.0), (2.0, 0.0), (2.0, 2.0), (0.0, 2.0)]),
            shapely.Polygon(
                shell=[
                    (0.0, 2.0),
                    (2.0, 2.0),
                    (2.0, 0.0),
                    (3.0, 0.0),
                    (3.0, 3.0),
                    (0.0, 3.0),
                ]
            ),
        ]
    )
    return polygons, values


def test_locate_polygon(grid):
    polygon = shapely.Polygon(shell=[(0.5, 0.5), (2.5, 0.5), (0.5, 2.5)])
    exterior = shapely.get_coordinates(polygon.exterior)
    interiors = [shapely.get_coordinates(interior) for interior in polygon.interiors]

    actual = burn._locate_polygon(grid, exterior, interiors, all_touched=False)
    assert np.array_equal(np.sort(actual), [0, 1, 2, 3, 4, 6])
    actual = burn._locate_polygon(grid, exterior, interiors, all_touched=True)
    assert np.array_equal(np.sort(actual), [0, 1, 2, 3, 4, 6])

    polygon = shapely.Polygon(shell=[(0.75, 0.5), (2.5, 0.5), (0.75, 2.5)])
    exterior = shapely.get_coordinates(polygon.exterior)
    interiors = [shapely.get_coordinates(interior) for interior in polygon.interiors]

    actual = burn._locate_polygon(grid, exterior, interiors, all_touched=False)
    assert np.array_equal(np.sort(actual), [1, 2, 4])
    actual = burn._locate_polygon(grid, exterior, interiors, all_touched=True)
    assert np.array_equal(np.sort(actual), [0, 1, 2, 3, 4, 5, 6, 7])


def test_locate_polygon_with_hole(grid):
    # The hole omits the centroid at (1.5, 1.5)
    polygon = shapely.Polygon(
        shell=[(0.7, 0.7), (2.3, 0.7), (1.5, 2.3)],
        holes=[[(1.4, 1.6), (1.5, 1.4), (1.6, 1.6)]],
    )
    exterior = shapely.get_coordinates(polygon.exterior)
    interiors = [shapely.get_coordinates(interior) for interior in polygon.interiors]

    actual = burn._locate_polygon(grid, exterior, interiors, all_touched=False)
    assert np.array_equal(actual, [])
    actual = burn._locate_polygon(grid, exterior, interiors, all_touched=True)
    assert np.array_equal(np.unique(actual), [0, 1, 2, 3, 4, 5, 7])


def test_burn_polygons(grid, polygons_and_values):
    polygons, values = polygons_and_values
    output = np.full(grid.n_face, np.nan)
    burn._burn_polygons(polygons, grid, values, all_touched=False, output=output)
    expected = np.array([0, 0, 1, 0, 0, 1, 1, 1, 1])
    assert np.allclose(output, expected)


def test_burn_points(grid, points_and_values):
    points, values = points_and_values
    output = np.full(grid.n_face, -1.0)

    burn._burn_points(points, grid, values, output=output)
    expected = np.array([0, 1, -1, -1, -1, -1, -1, -1, 3])
    assert np.allclose(output, expected)


def test_burn_lines(grid, lines_and_values):
    lines, values = lines_and_values
    output = np.full(grid.n_face, -1.0)

    burn._burn_lines(lines, grid, values, output=output)
    expected = np.array([0, 0, 0, -1, 1, -1, 2, 2, -1])
    assert np.allclose(output, expected)


def test_burn_vector_geometry__errors(grid, points_and_values):
    with pytest.raises(TypeError, match="gdf must be GeoDataFrame"):
        xu.burn_vector_geometry(0, grid)

    points, values = points_and_values
    gdf = gpd.GeoDataFrame({"values": values}, geometry=points)
    with pytest.raises(TypeError, match="Like must be Ugrid2d, UgridDataArray"):
        xu.burn_vector_geometry(gdf, gdf)

    # This seems like the easiest way to generate a multi-polygon inside a
    # GeoDataFrame, since it won't initialize with a multi-polygon.
    p1 = Polygon([(0, 0), (1, 0), (1, 1)])
    p2 = Polygon([(0, 0), (1, 0), (1, 1), (0, 1)])
    p3 = Polygon([(2, 0), (3, 0), (3, 1), (2, 1)])
    gdf = gpd.GeoDataFrame({"values": [0, 0, 0]}, geometry=[p1, p2, p3]).dissolve(
        by="values"
    )
    with pytest.raises(
        TypeError, match="GeoDataFrame contains unsupported geometry types"
    ):
        xu.burn_vector_geometry(gdf, grid)


def test_burn_vector_geometry(
    grid, points_and_values, lines_and_values, polygons_and_values
):
    polygons, poly_values = polygons_and_values
    gdf = gpd.GeoDataFrame({"values": poly_values}, geometry=polygons)
    actual = xu.burn_vector_geometry(gdf, grid)
    assert isinstance(actual, xu.UgridDataArray)
    assert np.allclose(actual.to_numpy(), 1)
    actual = xu.burn_vector_geometry(gdf, grid, all_touched=True)
    assert np.allclose(actual.to_numpy(), 1)

    expected = np.array([0, 0, 1, 0, 0, 1, 1, 1, 1])
    actual = xu.burn_vector_geometry(gdf, grid, column="values")
    assert np.allclose(actual.to_numpy(), expected)

    points, point_values = points_and_values
    lines, line_values = lines_and_values
    line_values += 10
    point_values += 20
    values = np.concatenate([poly_values, line_values, point_values])
    geometry = np.concatenate(
        [polygons.to_numpy(), lines.to_numpy(), points.to_numpy()]
    )
    gdf = gpd.GeoDataFrame({"values": values}, geometry=geometry)
    actual = xu.burn_vector_geometry(gdf, grid, column="values")
    expected = np.array([20.0, 21.0, 10.0, 0.0, 11.0, 1.0, 12.0, 12.0, 23.0])
    assert np.allclose(actual.to_numpy(), expected)

    # All touched should give the same answer for this specific example.
    actual = xu.burn_vector_geometry(gdf, grid, column="values", all_touched=True)
    assert np.allclose(actual.to_numpy(), expected)


def test_earcut_triangulate_polygons():
    xy = np.array(
        [
            [0.0, 0.0],
            [1.0, 0.0],
            [1.0, 1.0],
            [0.0, 1.0],
            [0.0, 0.0],
        ]
    )
    hole = np.array(
        [
            [
                [0.25, 0.25],
                [0.75, 0.25],
                [0.75, 0.75],
                [0.25, 0.25],
            ]
        ]
    )
    polygon = shapely.polygons(xy, holes=hole)
    gdf = gpd.GeoDataFrame(data={"a": [10.0], "b": [20.0]}, geometry=[polygon])
    uda = xu.earcut_triangulate_polygons(polygons=gdf)
    assert isinstance(uda, xu.UgridDataArray)
    assert np.allclose(uda.to_numpy(), 0)
    assert uda.name is None

    uda = xu.earcut_triangulate_polygons(polygons=gdf, column="a")
    assert np.allclose(uda.to_numpy(), 10.0)
    assert uda.name == "a"

    uda = xu.earcut_triangulate_polygons(polygons=gdf, column="b")
    assert np.allclose(uda.to_numpy(), 20.0)
    assert uda.name == "b"
